#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

import numpy as np
import xarray as xr
import itertools
import random
import math
import inspect
import copy
import operator as op
from functools import reduce

from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete, decode_discrete_to_grid, encode_grid_indv_to_discrete

from neorl import WOA
from neorl import GWO
from neorl import PSO
from neorl import MFO
from neorl import HHO
from neorl import DE
from neorl import JAYA
from neorl import SSA
from neorl import ES

# Note: to incorporate additional algorithms into ensembles, the following needs to be done:
#     1. detect_algo needs to be updated to return appropriate name
#     2. clone_algo_obj needs to be updated to change the correct attribute in the dict
#     3. eval_algo_popnumber needs to be given some criteria for the minimum number of individuals
#        to run the evolution phase

class FitWrap:
    """class to track function calls"""
    def __init__(self, f):
        self.n = 0
        self.outs = []
        self.ins = []
        self.fxn = f
    def f(self, *inputs): 
        ans = self.fxn(*inputs)
        self.n += 1
        self.ins.append(inputs)
        self.outs.append(ans)
        return ans
    def reset(self):
        self.n = 0
        self.outs = []
        self.ins = []

def ncr(n, r):
    r = min(r, n-r)
    numer = reduce(op.mul, range(n, n-r, -1), 1)
    denom = reduce(op.mul, range(1, r+1), 1)
    return numer // denom


def raj_logfact(n):
    """Ramanujan approximation of log(n!)"""
    f1 = n*np.log(n)
    f2 = -n
    f3 = np.log(n*(1+4*n*(1+2*n)))/6
    f4 = np.log(np.pi)/2
    return f1 + f2 + f3 + f4

def detect_algo(obj):
    #simple function to detect object type given neorl
    # algorithm object
    if isinstance(obj, WOA):
        return 'WOA'
    elif isinstance(obj, GWO):
        return 'GWO'
    elif isinstance(obj, PSO):
        return 'PSO'
    elif isinstance(obj, MFO):
        return 'MFO'
    elif isinstance(obj, HHO):
        return 'HHO'
    elif isinstance(obj, DE):
        return 'DE'
    elif isinstance(obj, ES):
        return 'ES'
    elif isinstance(obj, SSA):
        return 'SSA'
    elif isinstance(obj, JAYA):
        return 'JAYA'
    raise Exception('%s algorithm object not recognized or supported'%obj)

max_algos = ['PSO', 'DE', 'ES', 'JAYA']#algos that change fitness function to make a maximum problem
min_algos = ['WOA', 'GWO', 'MFO', 'HHO', 'SSA']#algos that change fitness function to make a minimum problem

def wtd_remove(lst, ei, wts = None):
    #quick helper function to handle removing ei items from lst and returning them with
    #wts probability vector
    if wts is None:
        wts = [1/len(lst) for i in range(len(lst))]

    wts_sum = sum(wts)
    wts_checked = [a/wts_sum for a in wts] #correct for errors may come from rajmujan factorial

    #start here, caused by zeros in the weight
    indxs = np.random.choice(range(len(lst)), size=ei, p = wts_checked, replace = False)
    return [lst.pop(i) for i in reversed(sorted(indxs))], indxs

def clone_algo_obj(obj, nmembers, fit, bounds):
    # function to return a copy of an algorithm object with anumber of members given as
    # nmembers. This is to circumvent the error when the x0 passed in the evolute function
    # is a different size than the individuals given originally in the initialization of 
    # the algorithm object.
    # works for now, has potential of breaking

    def filter_kw(dftr, twk):
        sig = inspect.signature(twk)
        fks = [p.name for p in sig.parameters.values() if p.kind == p.POSITIONAL_OR_KEYWORD]
        fd = {fk:dftr[fk] for fk in fks if fk in dftr.keys()}
        return fd

    algo = detect_algo(obj)

    attrs = obj.__dict__

    if algo == 'WOA':
        attrs['nwhales'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return WOA(**filter_kw(attrs, WOA))
    elif algo == 'GWO':
        attrs['nwolves'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return GWO(**filter_kw(attrs, GWO))
    elif algo == 'PSO':
        attrs['npar'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return PSO(**filter_kw(attrs, PSO))
    elif algo == 'HHO':
        attrs['nhawks'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return HHO(**filter_kw(attrs, HHO))
    elif algo == 'MFO':
        attrs['nmoths'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return MFO(**filter_kw(attrs, MFO))
    elif algo == 'DE':
        attrs['npop'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return DE(**filter_kw(attrs, DE))
    elif algo == 'ES':
        attrs['lambda_'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return ES(**filter_kw(attrs, ES))
    elif algo == 'SSA':
        attrs['nsalps'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return SSA(**filter_kw(attrs, SSA))
    elif algo == 'JAYA':
        attrs['npop'] = nmembers
        attrs['fit'] = fit
        attrs['bounds'] = bounds
        return JAYA(**filter_kw(attrs, JAYA))


def get_algo_nmembers(obj):
    # function to retrieve the number of 'members' in the starting population of an
    # algorithm object plassed as obj

    algo = detect_algo(obj)

    if algo == 'WOA':
        return obj.nwhales
    elif algo == 'GWO':
        return obj.nwolves
    elif algo == 'PSO':
        return obj.npar
    elif algo == 'HHO':
        return obj.nhawks
    elif algo == 'DE':
        return obj.npop
    elif algo == 'ES':
        return obj.lambda_
    elif algo == 'MFO':
        return obj.npop
    elif algo == 'SSA':
        return obj.nsalps
    elif algo == 'JAYA':
        return obj.npop

def get_algo_ngtonevals(obj):
    # function to retrieve the number of function evaluations
    # for a given algo, args are num of generations and number of individuals, 
    # returns number of fxn evaluations
    algo = detect_algo(obj)

    if algo in ['WOA', 'GWO', 'MFO']:
        return lambda i, a : a*i
    elif algo == 'PSO':
        return lambda i, a : (i+1)*a
    elif algo == 'DE':
        return lambda i, a : 2*i*a
    elif algo == 'ES':
        return lambda i, a : (i+1)*a
    elif algo == 'HHO':
        print('--warning: HHO uses an upper bound for number of function evaluations,'
        ' be careful when using HHO with burdened variants')
        return lambda i, a : 2*i*a
    elif algo == 'SSA':
        return lambda i, a : (i+1)*a
    elif algo == 'JAYA':
        return lambda i, a : (i+1)*a

def eval_algo_popnumber(obj, nmembers):
    # check if an algorithm is prepared to participate in evolution phase
    # based on its population information

    algo = detect_algo(obj)
    if algo in ['WOA', 'GWO', 'PSO', 'HHO', 'DE', 'SSA']:
        return nmembers >= 5
    elif algo in ['MFO', 'JAYA']:
        return nmembers >= 4
    elif algo == 'ES':
        return nmembers >= obj.mu

def get_algo_annealed_kwargs(obj, ncyc, Ncyc, gen_per_cycle):
    total_parm = Ncyc*gen_per_cycle
    start = (ncyc - 1)*gen_per_cycle/(total_parm - 1)
    stop = ((ncyc - 1)*gen_per_cycle + gen_per_cycle - 1)/(total_parm - 1)
    fracaneal = np.linspace(start, stop, gen_per_cycle)

    algo = detect_algo(obj)
    #for linear annealed parameters is start + (stop - start)*fracaneal
    if algo == 'GWO':
        k = {'a' : 2 + (0 - 2)*fracaneal}
    elif algo == 'HHO':
        k = {'E1' : 2 + (0 - 2)*fracaneal}
    elif algo == 'PSO':
        if obj.speed_mech == 'timew':
            k = {'w' : obj.wmax + (obj.wmin - obj.wmax)*fracaneal}
        else:
            k = {}
    elif algo == 'WOA':
        k = {'fac' : -1 + (-2 - -1)*fracaneal,
             'a' : 2 + (0 - 2)*fracaneal}
    elif algo == 'MFO':
        k = {'r' : -1 + (-2 - -1)*fracaneal}
    elif algo == 'SSA':
        k = {'c1' : 2*np.exp(-(4*fracaneal)**2)}
    else:
        k = {}

    return k


class Population:
    # Class to store information and functionality related to a single population
    # in the AEO algorithm. Should be characterized by an evolution strategy passed
    # as one of the optimization classes from the other algorithms in NEORL.
    def __init__(self, strategy, algo, init_pop, mode, conv = None):
        # strategy should be one of the optimization objects containing an "evolute" method
        # init_pop needs to match the population given to the strategy object initially
        # algo is string that identifies which class is being used
        # conv is a function which takes ngen and returns number of evaluations
        self.conv = conv
        self.strategy = copy.deepcopy(strategy)
        self.algo = algo
        self.members = init_pop
        self.n = len(self.members)
        self.mode = mode

        self.popname = '' #will be assigned externally after object has been initialized
                          #    used only for loggin purposes

        self.fitlog = []

    @property
    def fitness(self):
        return self.fitlog[-1]

    @property
    def dfitness(self):
        if len(self.fitlog) == 1:
            return self.fitlog[-1]
        else:
            return self.fitlog[-1] - self.fitlog[-2]

    def evolute(self, ngen, fit, bounds, ncyc, Ncyc, var_type, bounds_map, tbounds, log): #fit is included to avoid needing to pull .fit methods
                                       #    from algo objects that may have been flipped
        #check if there are enouh members to evolve NOT appended to fitlog
        if not eval_algo_popnumber(self.strategy, len(self.members)):
            if len(self.fitlog) == 0:
                raise Exception("Starting population for %s too small for evolution"%self.algo)
            fitness = self.fitness
            self.Nc = 0

        else:
            np.put(log['evolute'].data, [0], True)

            #update strategy with new population number
            self.strategy = clone_algo_obj(self.strategy, len(self.members), fit, bounds)

           #store last generation number
            self.last_ngen = ngen

            #perform evolution and store relevant information
            annealing_kwargs = get_algo_annealed_kwargs(self.strategy, ncyc, Ncyc, ngen)

            out = self.strategy.evolute(ngen, x0 = self.members, **annealing_kwargs)
            self.members = out[2]['last_pop'].iloc[:, :-1].values.tolist()
            self.member_fitnesses = out[2]['last_pop'].iloc[:, -1].values.tolist()

            self.fitlog.append(min(self.member_fitnesses))
            self.Nc = self.conv(self.last_ngen, self.n)

            fitness = self.fitness

        #log relevant information
        if self.n != 0:
            if 'grid' in var_type:
                p2 = []
                for m in self.members:
                    p2.append(encode_grid_indv_to_discrete(m, bounds, bounds_map))
                log['member_x'][:self.n] = np.array(p2).reshape(self.n, -1)
            else:
                log['member_x'][:self.n] = np.array(self.members).reshape(self.n, -1)
            log['member_fitnesses'][:self.n] = self.member_fitnesses
        np.put(log['nmembers'].data, [0],  [self.n])
        np.put(log['Nc'].data, [0], self.Nc)
        np.put(log['f'].data, [0], [self.fitness])
        if len(self.fitlog) > 1:
            np.put(log['delta_f'].data, [0], [self.dfitness])

        return fitness

    def strength(self, g, g_burden, fselmax, fselmin, log, g_or_b):
        #normalize strength measure
        if g == "improve":
            fsel = self.dfitness
        else:
            fsel = self.fitness

        normed = (fsel - fselmax)/(fselmin - fselmax)

        if g_or_b == 'g':
            np.put(log['unburdened_g'].data, [0], [normed])
        elif g_or_b == 'b':
            np.put(log['unburdened_b'].data, [0], [normed])

        #burden, if necessary
        if g_burden:
            normed /= 1 + self.Nc

        ret = normed

        if g_or_b == 'g':
            np.put(log['g'].data, [0], [ret])
        elif g_or_b == 'b':
            np.put(log['b'].data, [0], [ret])
        return ret

    def export(self, ei, wt, order, gfrac, log):
        #decide what members to export and then remove them from members and return them
        if wt == 'uni': #uniform case very easy
            shuffled = list(range(len(self.members)))
            if self.n > 0:
                removed_x, removed_idcs =  wtd_remove(self.members, ei)
            else:
                removed_x, removed_idcs = [], []
            np.put(log['wb'].data, [0], [False])
            if self.n > 0:
                log['export_wts'][:self.n] = 1/self.n
        else:
            if order[0] == 'a': #handle the annealed cases
                if gfrac < 0.5:
                    o = order[1:]
                else:
                    o = order[2] + order[1]
            else: #if no annealing
                o = order

            #order the members and note how they are shuffled
            shuffled = list(range(len(self.members)))
            if o == 'bw':
                self.members = [a for _, a in sorted(zip(self.member_fitnesses, self.members))]
                shuffled = [a for _, a in sorted(zip(self.member_fitnesses, shuffled))]
                self.member_fitnesses = sorted(self.member_fitnesses)
            if o == 'wb':
                self.members = [a for _, a in sorted(zip(self.member_fitnesses, self.members), reverse = True)]
                shuffled = [a for _, a in sorted(zip(self.member_fitnesses, shuffled), reverse = True)]
                self.member_fitnesses = sorted(self.member_fitnesses, reverse = True)

            #calculate the wts
            seq = np.array(range(1, len(self.members) + 1))
            if wt == 'log':
                wts = (np.log(seq)+1)/(self.n + raj_logfact(self.n))
            elif wt == 'lin':
                wts = (seq)/(self.n*.5+.5*self.n**2)
            elif wt == 'exp':
                wts = (np.exp(seq-1))/((1-np.exp(self.n))/(1-np.exp(1)))

            #log information
            np.put(log['wb'].data, [0], [o == 'wb'])
            log['export_wts'][:self.n][shuffled] = wts

            #draw members and return them
            if self.n > 0:
                removed_x, removed_idcs = wtd_remove(self.members, ei, wts)
            else:
                removed_x, removed_idcs = [], []

        np.put(log['exported'].data, [shuffled[j] for j in removed_idcs], [True]*len(removed_idcs))

        #remove member fitnesses to make sure len(members) == len(member_fitnesses)
        [self.member_fitnesses.pop(j) for j in reversed(sorted(removed_idcs))]

        self.n = len(self.members)

        #log which members were exported
        return removed_x

    def receive(self, individuals):
        #bring individuals into the populations
        self.members += individuals
        [self.member_fitnesses.append(np.nan) for i in individuals]
        self.n = len(self.members)


class AEO(object):
    """
    Animorphoc Ensemble Optimizer

    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function
    :param optimizers: (list) list of optimizer instances to be included in the ensemble
    :param gen_per_cycle: (int) number of generations performed in evolution phase per cycle
    :param mode: (str) problem type, either "min" for minimization problem or "max" for maximization
    :param seed: (int) random seed for sampling
    """
#    :param config: (int) If none, use migration parameters defined later, if int (1 through 3), use one of the presets
#    :param alpha: (float or str) option for exponent on ``g`` strength measure, if numeric, ``alpha`` is taken to be
#        that value. If ``alpha`` is "up" ``alpha`` is annealed from 0 to 1. If ``alpha`` is "down" it is annealed from
#        1 to 0.
#    :param g: (str) either "fitness" or "improve" for strength measure for exportation number section of migration
#    :param g_burden: (bool) True if strength if divided by number of fitness evaluations in evolution phase
#    :param q: (float or str) option for favoring weak or strong pops in exportation number
#    :param wt: (str) "log", "lin", "exp", "uni" for different weightings in member selection section of migration
#    :param beta: (float or str) option for exponent on ``b`` strength measure. See ``alpha`` for details.
#    :param b: (str) either "fitness" or "improve" for strength measure for destination selection section of migration
#    :param b_burden: (bool) True if strength if divided by number of fitness evaluations in evolution phase
#    :param order: (str) "wb" for worst to best, "bw" for best to worst, prepend "a" for annealed starting in the given ordering.
#    :param ncores: (int) number of parallel processors
    def __init__(self, bounds, fit,
            optimizers, gen_per_cycle, mode = "min", seed = None, **kwargs):

        config = kwargs.get("config", 1)
        alpha = kwargs.get("alpha", "up")
        g = kwargs.get("g", "improve")
        g_burden = kwargs.get("g_burden", False)
        q = kwargs.get("q", "up")
        wt = kwargs.get("wt", "exp")
        beta = kwargs.get("beta", "up")
        b = kwargs.get("b", "improve")
        b_burden = kwargs.get("b_burden", False)
        order = kwargs.get("order", "bw")

        if config is None:
            pass
        elif config == 1:
            alpha = 'up'
            g = 'improve'
            g_burden = False
            q = 'up'
            wt = 'exp'
            beta = 'up'
            b = 'improve'
            b_burden = False
        elif config == 2:
            alpha = 'up'
            g = 'improve'
            g_burden = False
            q = 1.
            wt = 'log'
            beta = 'up'
            b = 'improve'
            b_burden = True
        elif config == 3:
            alpha = 0.
            g = 'improve'
            g_burden = False
            q = -1.
            wt = 'exp'
            beta = 'up'
            b = 'fitness'
            b_burden = True
        else:
            raise Exception("Not approved configuration, select 1, 2 or 3")

        if not (seed is None):
            random.seed(seed)
            np.random.seed(seed)

        self.mode=mode
        self.wrapped_f = FitWrap(fit)
        self.fit = self.wrapped_f.f

        if mode == 'max': #create fit attribute to use for checking consistency of fits
            raise Exception("Max not supported for AEO")
        elif mode == 'min':
            self.fitcheck=fit
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')

        self.optimizers = optimizers
        self.algos = [detect_algo(o) for o in self.optimizers]
        self.gpc = gen_per_cycle

        self.bounds = bounds
        self.var_type = np.array([bounds[item][0] for item in bounds])

        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            self.trans_bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.trans_bounds = copy.copy(self.bounds)

        #get functions to convert number of generations to number of evaluaions
        self.ngtonevals = [get_algo_ngtonevals(a) for a in self.optimizers]

        #infer variable types
        self.var_type = np.array([bounds[item][0] for item in bounds])

        self.dim = len(bounds)

        #check that all optimizers have options that match AEO
        self.ensure_consistency()

        #process variant options for exportation number
        self.alpha = alpha
        if (not isinstance(self.alpha, float) and
            not self.alpha in ['up', 'down']):
            raise Exception('invalid value for alpha, make sure it is a float!')
        if isinstance(self.alpha, float):
            if self.alpha < 0:
                raise Exception('alpha must be equal to or greater than 0')

        self.g = g
        if not self.g in ['fitness', 'improve']:
            raise Exception('invalid option for g')

        self.g_burden = g_burden
        if not isinstance(g_burden, bool):
            raise Exception('g_burden should be boolean type')

        self.q = q
        if (not isinstance(self.q, float)) and not self.q in ['up', 'down']:
            raise Exception('invalid value for q, make sure it is float or string option')

        #process variant options for member selection
        self.wt = wt
        if not self.wt in ['log', 'lin', 'exp', 'uni']:
            raise Exception('invalid option for wt')

        self.order = order
        if not self.order in ['wb', 'bw', 'awb', 'abw', None]:
            raise Exception('invalid option for order')

        if self.wt == 'uni' and (self.order is not None):
            print('--warning: order options ignored for uniform weighting')

        #process variant options for destination selection
        self.beta = beta
        if (not isinstance(self.beta, float) and
            not self.beta in ['up', 'down']):
            raise Exception('invalid value for beta, make sure it is a float!')
        if isinstance(self.beta, float):
            if self.beta < 0:
                raise Exception('beta must be equal to or greater than 0')

        self.b = b
        if not self.b in ['fitness', 'improve']:
            raise Exception('invalid option for b')

        self.b_burden = b_burden
        if not isinstance(b_burden, bool):
            raise Exception('b_burden should be boolean type')

    def ensure_consistency(self):
        #loop through all optimizers and make sure all options are set to be the same
        return
        gen_warning = ', check that options of all optimizers are the same as AEO'
        for o, a in zip(self.optimizers, self.algos):
            assert self.mode == o.mode,'%s has incorrect optimization mode'%o + gen_warning
            assert self.bounds == o.bounds,'%s has incorrect bounds'%o + gen_warning

    def init_sample(self, bounds):
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')
        return indv

    def get_alphabeta(self, aorb, ncyc, Ncyc):
        if isinstance(aorb, float):
            return aorb
        elif aorb == 'up':
            return (ncyc-1)/(Ncyc-1)
        elif aorb == 'down':
            return 1 - (ncyc-1)/(Ncyc-1)

    def get_q(self, ncyc, Ncyc):
        if isinstance(self.q, float):
            return self.q
        elif self.q == "up":
            return 2/(1 - Ncyc)*(1 - ncyc)-1
        elif self.q == "down":
            return 2/(1 - Ncyc)*(ncyc - 1)+1

    def fill_M(self, log):
        h = log["export_pop_wts"].values
        scale_g = log["str_dest_scaled"].values
        I = log.coords["pop"].size
        nmems = log["nmembers"].values
        for i in range(I):
            for j in range(I):
                tot = 0
                for k in range(nmems[i] + 1):
                    tot += k/nmems[i]*ncr(nmems[i],k)*h[i]**k*(1-h[i])**(nmems[i]-k)
                kd = int(i==j) #kroniker delta function
                if nmems[i] == 0:
                   log['M'].values[j,i] = 1/len(nmems)#kd
                else:
                    log['M'].values[j,i] = kd + (scale_g[j] - kd)*tot

    def evolute(self, Ncyc, npop0 = None, x0 = None, pop0 = None, stop_criteria = None, verbose = False):
        """
        This function evolutes the AEO algorithm for a number of cycles. Either
        (``npop0``) or (``x0`` and ``pop0``) are required.

        :param Ncyc: (int) number of cycles to evolute
        :param pop0: (list of ints) number of individuals in starting population for each optimizer
        :param x0: (list of lists) initial positions of individuals in problem space
        :param pop0: (list of ints) population assignments for ``x0``, integer corresponding to assigned population ordered
            according to self.optimize
        :param stop_criteria: (None or callable) function which returns condition if evolution should continue, can be
            used to stop evolution at certain number of function evaluations

        :return: (tuple) (best individual, best fitness, xarray.Dataset of various algorithm parameters)
        """
        #if npop0, x0 and pop0 are none, detect populations from algos
        if npop0 is None and x0 is None and pop0 is None:
            npop0 = [get_algo_nmembers(a) for a in self.optimizers]
        #intepret npop0 or x0 and pop0 input
        if x0 is not None:
            if npop0 is not None:
                print('--warning: x0 and npop0 is defined, ignoring npop0')
            assert len(x0) == len(pop0), 'x0 and pop0 must be ov equal length'
        else:
            x0 = [self.init_sample(self.trans_bounds) for i in range(sum(npop0))]
            if 'grid' in self.var_type:
                x0_encode = copy.copy(x0)
                x0 = [decode_discrete_to_grid(a, self.bounds, self.bounds_map) for a in x0]
            dup = [[i]*npop0[i] for i in range(len(npop0))]
            pop0 = list(itertools.chain.from_iterable(dup))

        #separate starting positions according to optimizer/strategy, initialize Population objs
        self.pops = []
        for i in range(len(self.optimizers)):
            xpop = []
            for x, p in zip(x0, pop0):
                if p == i:
                    xpop.append(x)
            self.pops.append(Population(self.optimizers[i], self.algos[i],
                    xpop, self.mode, self.ngtonevals[i]))

        #initialize log Dataset
        membercoords = range(len(x0))
        cyclecoords = range(1, Ncyc + 1)
        popcoords = []
        for p in self.pops:
            algo = p.algo
            if not (algo in popcoords):
                p.popname = algo
                popcoords.append(p.popname)
            else:
                i = 2
                while algo + str(i).zfill(3) in popcoords:
                    i += 1
                p.popname = algo + str(i).zfill(3)
                popcoords.append(p.popname)

        varcoords = list(self.bounds.keys())

        nm = len(membercoords)
        nc = len(cyclecoords)
        npp = len(popcoords)
        nv = len(varcoords)



        log = xr.Dataset(
                {
                    'initial_member_x'   : (['member', 'pop',          'var'], np.zeros((nm, npp,     nv), dtype = np.float64)),
                    'member_x'           : (['member', 'pop', 'cycle', 'var'], np.zeros((nm, npp, nc, nv), dtype = np.float64)),
                    'nmembers'           : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.int32)),
                    'member_fitnesses'   : (['member', 'pop', 'cycle'], np.zeros((nm, npp, nc), dtype = np.float64)),
                    'nexport'            : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.int32)),
                    'export_str_scaled'  : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'export_pop_wts'     : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'alpha'              : ([                 'cycle'], np.zeros(          nc , dtype = np.float64)),
                    'wb'                 : ([                 'cycle'], np.zeros(          nc , dtype = np.bool8)),
                    'g'                  : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'f'                  : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'unburdened_g'       : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'Nc'                 : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.int32)),
                    'delta_f'            : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'fmin'               : ([                 'cycle'], np.zeros(          nc , dtype = np.float64)),
                    'fmax'               : ([                 'cycle'], np.zeros(          nc , dtype = np.float64)),
                    'dfmin'               : ([                 'cycle'], np.zeros(          nc , dtype = np.float64)),
                    'dfmax'               : ([                 'cycle'], np.zeros(          nc , dtype = np.float64)),
                    'migration'          : ([                 'cycle'], np.zeros(          nc , dtype = np.bool8)),
                    'export_wts'         : (['member', 'pop', 'cycle'], np.zeros((nm, npp, nc), dtype = np.float64)),
                    'exported'           : (['member', 'pop', 'cycle'], np.zeros((nm, npp, nc), dtype = np.bool8)),
#                    'pop_after_migrate'  : (['member', 'pop', 'cycle'], np.zeros((nm, npp, nc), dtype = '<U6')),
                    'beta'               : ([                 'cycle'], np.zeros(          nc , dtype = np.float64)),
                    'b'                  : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'unburdened_b'       : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'A'                  : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.int32)),
                    'M'                  : (['pop','popdest', 'cycle'], np.zeros((npp,npp, nc), dtype = np.float64)),
                    'str_dest_scaled'    : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.float64)),
                    'evolute'            : ([          'pop', 'cycle'], np.zeros((    npp, nc), dtype = np.bool8))},
                coords = {
                    'member'  : membercoords,
                    'pop'     : popcoords,
                    'cycle'   : cyclecoords,
                    'var'     : varcoords,
                    'popdest'     : popcoords}, #only used for migration transition matrix
                attrs = {"Ncycles" : 0}
                )
        #initial_member_x: positions of all individuals in a population before any evolution
        #member_x: positions of members in each population as through each cycle
        #nmembers: number of members in each population in each cycle
        #export_str_scaled: scaled strengths of each population each cycle
        #export_pop_wts: weights passed into binomial function for ei selection, called hi in paper
        #alpha: alpha parameter as it may change over each cycle
        #wb: ordering of individuals within the populations when used for member selection
        #g: unscaled strengths of each population each cycle
        #f: fitness of each population each cycle, may have existed across previous cycles
        #unburdened_g: unscaled strengths of each pop each cycle without the burden applied
        #Nc: Number of function evaluations required to get the f variable associated with that cycle
        #delta_f: difference in current cycle fitness to previous cycle fitness
        #fmin: minimum fitness across all populations for a single cycle
        #fmax: maximum fitness across all populations for a single cycle
        #dfmin: minimum improvement across all populations for a single cycle
        #dfmax: maximum improvement across all populations for a single cycle
        #migration: bool as to whether or not migration is performed, if fmax=fmix, no evolution
        #export_wts: weights used for each individual in the member selection phase
        #exported: boolean as to whether or not an individual was exported
        #beta: beta parameter as it may change over each cycle
        #b: unscaled strengths of each population for the destination selection phase
        #unburdened_b: unscaled strengths without burdened applied for member selection phase
        #A: number of members from export pool that end up in a particulat population each cycle
        #M: stochastic matrix which describes movement of members in a migration phase
        #evolute: is whether or not that population was evoluted each cycle



        #log positions of initial members

        if not 'grid' in self.var_type:
            for p in self.pops:
                log['initial_member_x'].loc[{'pop' : p.popname}][:len(p.members)] = p.members

        #perform evolution/migration cycle
        for i in range(1, Ncyc + 1):
            #evolution phase
            if "grid" in self.var_type:
                pop_fits = [p.evolute(self.gpc, self.fit, self.bounds, i, Ncyc, self.var_type, self.bounds_map, self.trans_bounds, log.loc[{'pop' : p.popname, 'cycle' : i}]) for p in self.pops]
            else:
                pop_fits = [p.evolute(self.gpc, self.fit, self.bounds, i, Ncyc, self.var_type, None, self.trans_bounds, log.loc[{'pop' : p.popname, 'cycle' : i}]) for p in self.pops]

            #exportation number
            #  calc weights
            maxf = max(pop_fits)
            minf = min(pop_fits)
            maxdf = max([p.dfitness for p in self.pops])
            mindf = min([p.dfitness for p in self.pops])
            alpha = self.get_alphabeta(self.alpha, i, Ncyc)
            if (maxf == minf and self.g == "fitness") or (maxdf == mindf and self.g == "improve"):#export nobody if this true
                eis = [0]*len(self.pops)
                log['migration'].loc[{'cycle' : i}] = False
            else:
                if self.g == "fitness":
                    strengths_exp = [p.strength(self.g, self.g_burden, maxf, minf, log.loc[{'pop' : p.popname, 'cycle' : i}], 'g')**alpha for p in self.pops]
                else:
                    strengths_exp = [p.strength(self.g, self.g_burden, maxdf, mindf, log.loc[{'pop' : p.popname, 'cycle' : i}], 'g')**alpha for p in self.pops]
                strengths_exp_scaled = [s/sum(strengths_exp) for s in strengths_exp]

                #  sample binomial to get e_i for each population
                qt = self.get_q(i, Ncyc)
                binomial_wts = [(.5 - s)*qt + .5 for s in strengths_exp_scaled]
                eis = [np.random.binomial(len(p.members), binomial_wts[j]) for j, p in enumerate(self.pops)]
                log['migration'].loc[{'cycle' : i}] = True

                log['export_str_scaled'].loc[{'cycle' : i}] = strengths_exp_scaled
                log['export_pop_wts'].loc[{'cycle' : i}] = binomial_wts

            # log pop export info
            log['nexport'].loc[{'cycle' : i}] = eis
            log['fmax'].loc[{'cycle' : i}] = maxf
            log['fmin'].loc[{'cycle' : i}] = minf
            log['dfmax'].loc[{'cycle' : i}] = maxdf
            log['dfmin'].loc[{'cycle' : i}] = mindf
            log['alpha'].loc[{'cycle' : i}] = alpha

            #member selection
            #  members removed from population with this export method
            exported = [p.export(eis[j], self.wt, self.order, i/Ncyc, log.loc[{'pop' : p.popname, 'cycle' : i}]) for j, p in enumerate(self.pops)]

            #destination selection
            beta = self.get_alphabeta(self.beta, i, Ncyc)
            log['beta'].loc[{'cycle' : i}] = beta

            if (maxf == minf and self.b == "fitness") or (maxdf == mindf and self.b == "improve"): #doesn't matter because nobody is exported
                strengths_dest = [1.]*len(self.pops)
            else:
                if self.b == "fitness":
                    strengths_dest = [p.strength(self.b, self.b_burden, maxf, minf, log.loc[{'pop' : p.popname, 'cycle' : i}], 'b')**beta for p in self.pops]
                else:
                    strengths_dest = [p.strength(self.b, self.b_burden, maxdf, mindf, log.loc[{'pop' : p.popname, 'cycle' : i}], 'b')**beta for p in self.pops]

            #manage members that are currently without a home
            exported = list(itertools.chain.from_iterable(exported))
            random.shuffle(exported)#in-place randomize order

            #calculate normalized probabilities and draw samples
            strengths_dest_scaled = [s/sum(strengths_dest) for s in strengths_dest]
            allotments = np.random.multinomial(len(exported), strengths_dest_scaled)
            log['str_dest_scaled'].loc[{'cycle' : i}] = strengths_dest_scaled
            log['A'].loc[{'cycle' : i}] = allotments

            #distribute individuals according to the sample
            for a, p in zip(allotments, self.pops):
                p.receive(exported[:a])
                exported = exported[a:]

            #calculate migration matrix
            self.fill_M(log.loc[{'cycle' : i}])

            #check if desired number of function evaluations has been reached
            log.attrs["Ncycles"] = i
            if not stop_criteria is None:
                if stop_criteria() == True:
                    break
            
            #bestind = np.argmin(self.wrapped_f.outs)
            #print(self.wrapped_f.ins[bestind][0], self.wrapped_f.outs[bestind])
            #assert self.fit(self.wrapped_f.ins[bestind][0]) == self.wrapped_f.outs[bestind]
            #print(self.wrapped_f.outs[bestind])

        #get best members
        bestind = np.argmin(self.wrapped_f.outs)

        xbest = self.wrapped_f.ins[bestind]
        ybest = self.wrapped_f.outs[bestind]

        xbest_correct = xbest[0]

        return xbest_correct, ybest, log



from abc import ABC, abstractmethod
import typing
from typing import Union, Optional, Any

import gym
import numpy as np

from neorl.rl.baselines.shared.callbacks import BaseCallback
from neorl.rl.baselines.shared.vec_env import VecEnv

if typing.TYPE_CHECKING:
    from neorl.rl.baselines.shared.base_class import BaseRLModel  # pytype: disable=pyi-error


class AbstractEnvRunner(ABC):
    def __init__(self, *, env: Union[gym.Env, VecEnv], model: 'BaseRLModel', n_steps: int):
        """
        Collect experience by running `n_steps` in the environment.
        Note: if this is a `VecEnv`, the total number of steps will
        be `n_steps * n_envs`.

        :param env: (Union[gym.Env, VecEnv]) The environment to learn from
        :param model: (BaseRLModel) The model to learn
        :param n_steps: (int) The number of steps to run for each environment
        """
        self.env = env
        self.model = model
        n_envs = env.num_envs
        self.batch_ob_shape = (n_envs * n_steps,) + env.observation_space.shape
        self.obs = np.zeros((n_envs,) + env.observation_space.shape, dtype=env.observation_space.dtype.name)
        self.obs[:] = env.reset()
        self.n_steps = n_steps
        self.states = model.initial_state
        self.dones = [False for _ in range(n_envs)]
        self.callback = None  # type: Optional[BaseCallback]
        self.continue_training = True
        self.n_envs = n_envs

    def run(self, callback: Optional[BaseCallback] = None) -> Any:
        """
        Collect experience.

        :param callback: (Optional[BaseCallback]) The callback that will be called
            at each environment step.
        """
        self.callback = callback
        self.continue_training = True
        return self._run()

    @abstractmethod
    def _run(self) -> Any:
        """
        This method must be overwritten by child class.
        """
        raise NotImplementedError


def traj_segment_generator(policy, env, horizon, reward_giver=None, gail=False, callback=None):
    """
    Compute target value using TD(lambda) estimator, and advantage with GAE(lambda)
    :param policy: (MLPPolicy) the policy
    :param env: (Gym Environment) the environment
    :param horizon: (int) the number of timesteps to run per batch
    :param reward_giver: (TransitionClassifier) the reward predicter from obsevation and action
    :param gail: (bool) Whether we are using this generator for standard trpo or with gail
    :param callback: (BaseCallback)
    :return: (dict) generator that returns a dict with the following keys:
        - observations: (np.ndarray) observations
        - rewards: (numpy float) rewards (if gail is used it is the predicted reward)
        - true_rewards: (numpy float) if gail is used it is the original reward
        - vpred: (numpy float) action logits
        - dones: (numpy bool) dones (is end of episode, used for logging)
        - episode_starts: (numpy bool)
            True if first timestep of an episode, used for GAE
        - actions: (np.ndarray) actions
        - nextvpred: (numpy float) next action logits
        - ep_rets: (float) cumulated current episode reward
        - ep_lens: (int) the length of the current episode
        - ep_true_rets: (float) the real environment reward
        - continue_training: (bool) Whether to continue training
            or stop early (triggered by the callback)
    """
    # Check when using GAIL
    assert not (gail and reward_giver is None), "You must pass a reward giver when using GAIL"

    # Initialize state variables
    step = 0
    action = env.action_space.sample()  # not used, just so we have the datatype
    observation = env.reset()

    cur_ep_ret = 0  # return in current episode
    current_it_len = 0  # len of current iteration
    current_ep_len = 0  # len of current episode
    cur_ep_true_ret = 0
    ep_true_rets = []
    ep_rets = []  # returns of completed episodes in this segment
    ep_lens = []  # Episode lengths

    # Initialize history arrays
    observations = np.array([observation for _ in range(horizon)])
    true_rewards = np.zeros(horizon, 'float32')
    rewards = np.zeros(horizon, 'float32')
    vpreds = np.zeros(horizon, 'float32')
    episode_starts = np.zeros(horizon, 'bool')
    dones = np.zeros(horizon, 'bool')
    actions = np.array([action for _ in range(horizon)])
    states = policy.initial_state
    episode_start = True  # marks if we're on first timestep of an episode
    done = False

    callback.on_rollout_start()

    while True:
        action, vpred, states, _ = policy.step(observation.reshape(-1, *observation.shape), states, done)
        # Slight weirdness here because we need value function at time T
        # before returning segment [0, T-1] so we get the correct
        # terminal value
        if step > 0 and step % horizon == 0:
            callback.update_locals(locals())
            callback.on_rollout_end()
            yield {
                    "observations": observations,
                    "rewards": rewards,
                    "dones": dones,
                    "episode_starts": episode_starts,
                    "true_rewards": true_rewards,
                    "vpred": vpreds,
                    "actions": actions,
                    "nextvpred": vpred[0] * (1 - episode_start),
                    "ep_rets": ep_rets,
                    "ep_lens": ep_lens,
                    "ep_true_rets": ep_true_rets,
                    "total_timestep": current_it_len,
                    'continue_training': True
            }
            _, vpred, _, _ = policy.step(observation.reshape(-1, *observation.shape))
            # Be careful!!! if you change the downstream algorithm to aggregate
            # several of these batches, then be sure to do a deepcopy
            ep_rets = []
            ep_true_rets = []
            ep_lens = []
            # Reset current iteration length
            current_it_len = 0
            callback.on_rollout_start()
        i = step % horizon
        observations[i] = observation
        vpreds[i] = vpred[0]
        actions[i] = action[0]
        episode_starts[i] = episode_start

        clipped_action = action
        # Clip the actions to avoid out of bound error
        if isinstance(env.action_space, gym.spaces.Box):
            clipped_action = np.clip(action, env.action_space.low, env.action_space.high)

        if gail:
            reward = reward_giver.get_reward(observation, clipped_action[0])
            observation, true_reward, done, info = env.step(clipped_action[0])
        else:
            observation, reward, done, info = env.step(clipped_action[0])
            true_reward = reward

        if callback is not None:
            callback.update_locals(locals())
            if callback.on_step() is False:
                # We have to return everything so pytype does not complain
                yield {
                    "observations": observations,
                    "rewards": rewards,
                    "dones": dones,
                    "episode_starts": episode_starts,
                    "true_rewards": true_rewards,
                    "vpred": vpreds,
                    "actions": actions,
                    "nextvpred": vpred[0] * (1 - episode_start),
                    "ep_rets": ep_rets,
                    "ep_lens": ep_lens,
                    "ep_true_rets": ep_true_rets,
                    "total_timestep": current_it_len,
                    'continue_training': False
                    }
                return

        rewards[i] = reward
        true_rewards[i] = true_reward
        dones[i] = done
        episode_start = done

        cur_ep_ret += reward
        cur_ep_true_ret += true_reward
        current_it_len += 1
        current_ep_len += 1
        if done:
            # Retrieve unnormalized reward if using Monitor wrapper
            maybe_ep_info = info.get('episode')
            if maybe_ep_info is not None:
                if not gail:
                    cur_ep_ret = maybe_ep_info['r']
                cur_ep_true_ret = maybe_ep_info['r']

            ep_rets.append(cur_ep_ret)
            ep_true_rets.append(cur_ep_true_ret)
            ep_lens.append(current_ep_len)
            cur_ep_ret = 0
            cur_ep_true_ret = 0
            current_ep_len = 0
            if not isinstance(env, VecEnv):
                observation = env.reset()
        step += 1



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#"""
#Created on Wed Mar  4 11:51:22 2020
#
#@author: majdi
#"""

import numpy as np
import pandas as pd
from neorl.rl.baselines.shared.callbacks import BaseCallback
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import os
import copy

class SavePlotCallback(BaseCallback):
    """
    Callback for saving a model (the check is done every ``check_freq`` steps)
    based on the training reward (in practice, we recommend using ``EvalCallback``).
    """
    def __init__(self, check_freq, avg_step, log_dir, total_timesteps, basecall, plot_mode='subplot'):
        self.base=basecall
        self.plot_mode=plot_mode
        self.n_calls=self.base.n_calls
        self.model=self.base.model
        self.num_timesteps=self.base.num_timesteps
        self.total_timesteps=total_timesteps
        self.verbose=1
        self.check_freq = check_freq
        self.avg_step=avg_step
        self.log_dir = log_dir
        self.best_save_path = self.log_dir + '_bestmodel.pkl'
        self.save_path = self.log_dir + '_lastmodel.pkl'
        self.best_mean_reward = -np.inf

        #avoid activating 'Agg' in the header so not to affect other classes/algs
        import matplotlib
        matplotlib.use('Agg')

    def runcall(self):
        
        print('num_timesteps={}/{}'.format (self.num_timesteps, self.total_timesteps))
            
        # Retrieve training reward
        y= pd.read_csv(self.log_dir+'_out.csv')
        y=y["reward"].values
        # Mean training reward over the last 100 episodes
        mean_reward = np.mean(y[-self.avg_step:])
               
        # New best model, you could save the agent here
        print('--debug: current mean reward={}, previous best mean reward = {}'.format(np.round(mean_reward), np.round(self.best_mean_reward)))
        if mean_reward > self.best_mean_reward:
              self.best_mean_reward = copy.copy(mean_reward)
              #saving best model
              print('--debug: improvement in reward is observed, new best model is saved to {}'.format(self.best_save_path))
              self.model.save(self.best_save_path)    #best model found so far

        #saving current model
        print('--debug: current model model is saved to {}'.format(self.save_path))
        self.model.save(self.save_path)   #latest model
              
        self.out_data=pd.read_csv(self.log_dir+'_out.csv')
        #-------------------
        # Progress Plot
        #-------------------
        self.plot_progress()
                
    def _on_step(self) -> bool:
        
        try:
            if (self.num_timesteps % self.check_freq == 0) or (self.num_timesteps == self.total_timesteps):
                self.runcall()
        except:
            print('--warning: No plot is generated, NEORL tried to plot the output csv logger, but failed for some reason, you may increase `check_freq` to a large value to allow some data printed in the csv logger')
        
        if self.num_timesteps == self.total_timesteps:
            print('system exit')
            os._exit(1)
            
            
        return True
    
    def _on_training_end(self) -> None:
        self.runcall()
        print('Training is finished')
        os._exit(1)
        #pass

    def calc_cumavg(self, data, N):
    
        cum_aves=[np.mean(data[i:i+N]) for i in range(0,len(data),N)]
        cum_std=[np.std(data[i:i+N]) for i in range(0,len(data),N)]
        cum_max=[np.max(data[i:i+N]) for i in range(0,len(data),N)]
        cum_min=[np.min(data[i:i+N]) for i in range(0,len(data),N)]
    
        return cum_aves, cum_std, cum_max, cum_min
    
    
    def plot_progress(self, method_xlabel='Epoch'):

        self.out_data=pd.read_csv(self.log_dir+'_out.csv')
        color_list=['b', 'g', 'r', 'c', 'm', 'y', 'darkorange', 'purple', 'tab:brown', 'lime']
        plot_data=self.out_data.drop(['caseid'], axis=1)  #exclude caseid, which is the first column from plotting (meaningless)
        
        labels=list(plot_data.columns.values)
            
        ny=plot_data.shape[1] 
        
        assert ny == len(labels), 'number of columns ({}) to plot in the csv file {} is not equal to the number of labels provided by the user ({})'.format(ny, self.log_dir+'_out.csv', len(labels))
        
        # classic mode
        if self.plot_mode=='classic' or ny == 1:
            color_index=0
            for i in range (ny): #exclude caseid from plot, which is the first column 
                plt.figure()
                ravg, rstd, rmax, rmin=self.calc_cumavg(plot_data.iloc[:,i],self.avg_step)
                epochs=np.array(range(1,len(ravg)+1),dtype=int)
                plt.plot(epochs, ravg,'-o', c=color_list[color_index], label='Average per {}'.format(method_xlabel))
                
                plt.fill_between(epochs,[a_i - b_i for a_i, b_i in zip(ravg, rstd)], [a_i + b_i for a_i, b_i in zip(ravg, rstd)],
                alpha=0.2, edgecolor=color_list[color_index], facecolor=color_list[color_index], label=r'$1-\sigma$ per {}'.format(method_xlabel))
                
                plt.plot(epochs, rmax,'s', c='k', label='Max per {}'.format(method_xlabel), markersize=4)
                plt.plot(epochs,rmin,'d', c='k', label='Min per {}'.format(method_xlabel), markersize=4)
                plt.legend()
                plt.xlabel(method_xlabel)
                plt.ylabel(labels[i])
                
                if color_index==9:
                    color_index=0
                else:
                    color_index+=1
                    
                plt.tight_layout()
                plt.savefig(self.log_dir+'_'+labels[i]+'.png', format='png', dpi=150)
                plt.close()
        
        # subplot mode           
        elif self.plot_mode=='subplot':
            # determine subplot size
            if ny == 2:
                xx= [(1,2,1),(1,2,2)]
                plt.figure(figsize=(12, 4.0))
            elif ny==3:
                xx= [(1,3,1), (1,3,2), (1,3,3)]
                plt.figure(figsize=(12, 4.0))
            elif ny==4:
                xx= [(2,2,1), (2,2,2), (2,2,3), (2,2,4)]
                plt.figure(figsize=(12, 8))
            elif ny > 4 and ny <= 21:
                nrows=int(np.ceil(ny/3))
                xx= [(nrows,3,item) for item in range(1,ny+1)]
                adj_fac=(nrows - 2.0)*0.25 + 1
                plt.figure(figsize=(12, adj_fac*8))
            elif ny > 21 and ny <= 99:
                nrows=int(np.ceil(ny/4))
                xx= [(nrows,4,item) for item in range(1,ny+1)]
                adj_fac=(nrows - 2.0)*0.25 + 1
                plt.figure(figsize=(15, adj_fac*8))
                
                
            color_index=0
            for i in range (ny): #exclude caseid from plot, which is the first column 
                plt.subplot(xx[i][0], xx[i][1], xx[i][2])
                ravg, rstd, rmax, rmin=self.calc_cumavg(plot_data.iloc[:,i],self.avg_step)
                epochs=np.array(range(1,len(ravg)+1),dtype=int)
                plt.plot(epochs,ravg,'-o', c=color_list[color_index])
                
                plt.fill_between(epochs,[a_i - b_i for a_i, b_i in zip(ravg, rstd)], [a_i + b_i for a_i, b_i in zip(ravg, rstd)],
                alpha=0.2, edgecolor=color_list[color_index], facecolor=color_list[color_index])
                
                plt.plot(epochs,rmax,'s', c='k', markersize=4)
                
                plt.plot(epochs,rmin,'d', c='k', markersize=4)
                plt.xlabel(method_xlabel)
                plt.ylabel(labels[i])
                if color_index==9:
                    color_index=0
                else:
                    color_index+=1
            
            #speical legend is created for all subplots to save space
            legend_elements = [Line2D([0], [0], color='k', marker='o', label='Mean ' + r'$\pm$ ' +r'$1\sigma$' + ' per {} (color changes)'.format(method_xlabel)),
                  Line2D([0], [0], color='k', marker='s', label='Max per {} (color changes)'.format(method_xlabel)),
                  Line2D([0], [0], linestyle='-.', color='k', marker='d', label='Min per {} (color changes)'.format(method_xlabel))]
            plt.figlegend(handles=legend_elements, loc='upper center', bbox_to_anchor=(0.5, 1.02), ncol=3)
            plt.tight_layout()
            plt.savefig(self.log_dir+'_res.png', format='png', dpi=200, bbox_inches="tight")
            plt.close()
            
        else:
            raise Exception ('the plot mode defined by the user does not exist')
    
class RLLogger(BaseCallback):
    """
    Callback for logging data of RL algorathims (x,y), compatible with: A2C, ACER, ACKTR, DQN, PPO

    :param check_freq: (int) logging frequency, e.g. 1 will record every time step 
    :param plot_freq: (int) frequency of plotting the fitness progress (if ``None``, plotter is deactivated)
    :param n_avg_steps: (int) if ``plot_freq`` is NOT ``None``, then this is the number of timesteps to group to draw statistics for the plotter (e.g. 10 will group every 10 time steps to estimate min, max, mean, and std).
    :param pngname: (str) name of the plot that will be saved if ``plot_freq`` is NOT ``None``.
    :param save_model: (bool) whether or not to save the RL neural network model (model is saved every ``check_freq``)
    :param model_name: (str) name of the model to be saved  if ``save_model=True``
    :param save_best_only: (bool) if ``save_model = True``, then this flag only saves the model if the fitness value improves. 
    :param verbose: (bool) print updates to the screen
    """
    def __init__(self, check_freq=1, plot_freq=None, n_avg_steps=10, pngname='history', 
                 save_model=False, model_name='bestmodel.pkl', save_best_only=True, 
                 verbose=False):
        super(RLLogger, self).__init__(verbose)
        self.check_freq = check_freq
        self.plot_freq=plot_freq
        self.pngname=pngname
        self.n_avg_steps=n_avg_steps
        self.model_name = model_name
        self.save_model=save_model
        self.verbose=verbose
        self.save_best_only=save_best_only
        self.rbest = -np.inf
        self.rbest_maxonly = -np.inf
        self.r_hist=[]
        self.x_hist=[]
        
        if self.plot_freq:
            #avoid activating 'Agg' in the header so not to affect other classes/algs
            import matplotlib
            matplotlib.use('Agg')
            
    def _init_callback(self) -> None:
        # Create folder if needed
        try:
            self.mode=self.training_env.get_attr('mode')[0]   #PPO/ACER/A2C/ACKTR
        except:
            try:
                self.mode=self.training_env.mode       #DQN
            except:
                print('--warning: the logger cannot find mode in the environment, it is set by default to `max`')
                self.mode='max'
        
        if self.mode not in ['min', 'max']:
            self.mode='max'
            print('--warning: The mode entered by user is invalid, use either `min` or `max`')

        #if self.save_model:
        #    if self.log_dir is not None:
        #        os.makedirs(self.log_dir, exist_ok=True)

    def _on_step(self) -> bool:
        
        
        if self.n_calls % self.check_freq == 0:
            
            if self.verbose:
                print('----------------------------------------------------------------------------------')
                print('RL callback at step {}/{}'.format(self.n_calls, self.locals['total_timesteps']))
            
            try:
                rwd=self.locals['rew']   #DQN case (special dict naming)
            except:
                rwd=self.locals['rewards'][0] #A2C/PPO/ACER/ACKTR
                
            try:
                x=self.locals['infos'][0]['x'] #A2C/PPO/ACKTR cases
            except:
                if 'mus' in list(self.locals.keys()):
                    x=self.locals['_'][0]['x']     #ACER case (special dict naming)
                else:
                    x=self.locals['info']['x']   #DQN case (special dict naming)
                    
            if self.save_model and not self.save_best_only:
                self.model.save(self.model_name)
                if self.verbose:
                    print('A new model is saved to {}'.format(self.model_name))
                
            if rwd > self.rbest_maxonly:
                self.xbest=x.copy()
                self.rbest_maxonly=rwd
                
                if self.mode=='max':
                    self.rbest=self.rbest_maxonly
                else:
                    self.rbest=-self.rbest_maxonly
                
            
                if self.save_model and self.save_best_only:
                    self.model.save(self.model_name)
                    if self.verbose:
                        print('An improvement is observed, new model is saved to {}'.format(self.model_name))
            
            if self.mode=='max':
                self.r_hist.append(rwd)
            else:
                self.r_hist.append(-rwd)
            
            self.x_hist.append(list(x))
            
            if self.plot_freq:
                if self.n_calls % self.plot_freq == 0:
                    self.plot_progress()
                
            
            if self.verbose:
                print('----------------------------------------------------------------------------------')
        return True
    
    def plot_progress(self): 
    
        plt.figure()
        
        ravg, rstd, rmax, rmin=self.calc_cumavg(self.r_hist,self.n_avg_steps)
        epochs=np.array(range(1,len(ravg)+1),dtype=int)
        plt.plot(epochs, ravg,'-o', c='g', label='Average per epoch')
        
        plt.fill_between(epochs,[a_i - b_i for a_i, b_i in zip(ravg, rstd)], [a_i + b_i for a_i, b_i in zip(ravg, rstd)],
        alpha=0.2, edgecolor='g', facecolor='g', label=r'$1-\sigma$ per epoch')
        
        plt.plot(epochs, rmax,'s', c='k', label='Max per epoch', markersize=4)
        plt.plot(epochs,rmin,'d', c='k', label='Min per epoch', markersize=4)
        plt.legend()
        plt.xlabel('Epoch')
        plt.ylabel('Fitness')
        plt.savefig(self.pngname+'.png',format='png' ,dpi=300, bbox_inches="tight")
        plt.close()

    def calc_cumavg(self, data, N):
    
        cum_aves=[np.mean(data[i:i+N]) for i in range(0,len(data),N)]
        cum_std=[np.std(data[i:i+N]) for i in range(0,len(data),N)]
        cum_max=[np.max(data[i:i+N]) for i in range(0,len(data),N)]
        cum_min=[np.min(data[i:i+N]) for i in range(0,len(data),N)]
    
        return cum_aves, cum_std, cum_max, cum_min


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu March  3 14:42:29 2022
#
#@author: Majdi Radaideh
#"""

import numpy as np
from numpy import arange, multiply

def ensure_bounds(vec, bounds):

    vec_new = []
    # cycle through each variable in vector 
    for i, (key, val) in enumerate(bounds.items()):

        # variable exceedes the minimum boundary
        if vec[i] < bounds[key][1]:
            vec_new.append(bounds[key][1])

        # variable exceedes the maximum boundary
        if vec[i] > bounds[key][2]:
            vec_new.append(bounds[key][2])

        # the variable is fine
        if bounds[key][1] <= vec[i] <= bounds[key][2]:
            vec_new.append(vec[i])
        
    return vec_new

def generator(p,bounds,i,
              F,CR,popsize,
              dim,paraIndex,*args,**kwargs):

    #.... "rand/1/bin" strategy ....#
    
    u=np.zeros((3,dim))
    # Choose the indices for mutation
    indexSet=list(arange(popsize))
    indexSet.remove(indexSet[i])
    index=[]

    # Choose the first Index
    temp=int(np.floor(np.random.uniform() * (popsize - 2)) + 1); index.append(indexSet[temp]); indexSet.remove(indexSet[temp])

    # Choose the second index
    temp=int(np.floor(np.random.uniform() * (popsize - 3)) + 1); index.append(indexSet[temp]); indexSet.remove(indexSet[temp])
    
    # Choose the third index
    temp=int(np.floor(np.random.uniform() * (popsize - 4)) + 1); index.append(indexSet[temp])
    
    index=np.array(index)
    
    # Mutation
    v1=p[index[0],:] + F[paraIndex[0]] * (p[index[1],:] - p[index[2],:])

    # Handle the elements of the mutant vector which violate the boundary
    v1=ensure_bounds(v1, bounds)
        
    # Binomial crossover
    j_rand=int(np.floor(np.random.uniform() * dim))
    t=np.random.uniform(size=(dim)) < CR[paraIndex[0]]
    t[j_rand]=1
    t_=1 - t
    u[0,:]=multiply(t,v1) + multiply(t_,p[i,:])
    
    #... "current to rand/1" strategy ...#
    
    # The mechanism to choose the indices for mutation is slightly different from that of the classic
    # "current to rand/1", we found that using the following mechanism to choose the indices for
    # mutation can improve the performance to certain degree
    index[0:3]=np.floor(np.random.uniform(size=(3)) * popsize)
    
    index=index.astype(int)

    # Mutation
    v2=p[i,:] + np.random.uniform() * (p[index[0],:] - p[i,:]) + F[paraIndex[1]] * (p[index[1],:] - p[index[2],:])
    v2=ensure_bounds(v2, bounds)

    # Binomial crossover is not used to generate the trial vector under this condition
    u[1,:]=v2
    
    #... "rand/2/bin" strategy ...#
    
    # Choose the indices for mutation
    indexSet=list(arange(popsize))
    indexSet.remove(indexSet[i])
    index=[]

    # Choose the first index
    temp=int(np.floor(np.random.uniform() * (popsize - 2)) + 1); index.append(indexSet[temp]); indexSet.remove(indexSet[temp])

    # Choose the second index
    temp=int(np.floor(np.random.uniform() * (popsize - 3)) + 1); index.append(indexSet[temp]); indexSet.remove(indexSet[temp])

    # Choose the third index
    temp=int(np.floor(np.random.uniform() * (popsize - 4)) + 1); index.append(indexSet[temp]); indexSet.remove(indexSet[temp])
    
    # Choose the fourth index
    temp=int(np.floor(np.random.uniform() * (popsize - 5)) + 1); index.append(indexSet[temp]); indexSet.remove(indexSet[temp])

    # Choose the fifth index
    temp=int(np.floor(np.random.uniform() * (popsize - 6)) + 1); index.append(indexSet[temp])

    # Mutation
    # The first scaling factor (F) is randomly chosen from 0 to 1
    v3=p[index[0],:] + np.random.uniform() * (p[index[1],:] - p[index[2],:]) + F[paraIndex[2]] * (p[index[3],:]- p[index[4],:])

    # Handle the elements of the mutant vector which violate the boundary
    v3=ensure_bounds(v3, bounds)
    
    # Binomial crossover
    j_rand=int(np.floor(np.random.uniform() * dim))
    t=np.random.uniform(size=(dim)) < CR[paraIndex[2]]
    t[j_rand]=1
    t_=1 - t
    u[2,:]=multiply(t,v3) + multiply(t_,p[i,:])
    
    return u

def sub2ind(sz, row, col):
    n_rows = sz[0]
    return [n_rows * (c-1) + r for r, c in zip(row, col)]

def gnR1R2(NP1, NP2, r0,*args,**kwargs):
    # gnA1A2 generate two column vectors r1 and r2 of size NP1 & NP2, respectively
    #r1's elements are choosen from {1, 2, ..., NP1} & r1(i) ~= r0(i)
    #r2's elements are choosen from {1, 2, ..., NP2} & r2(i) ~= r1(i) & r2(i) ~= r0(i)
    
    # Call:
    #[r1 r2 ...] = gnA1A2(NP1)   # r0 is set to be (1:NP1)'
    #[r1 r2 ...] = gnA1A2(NP1, r0) # r0 should be of length NP1
    NP0=len(r0)
    r1=np.floor(np.random.uniform(size=(NP0))*NP1) + 1
    for i in range(1,10000):
        pos = (r1 == r0)
        if np.sum(pos) == 0:
            break
        else:
            r1[pos]=np.floor(np.random.uniform(size=(sum(pos))) * NP1) + 1
        if i > 1000:
            raise Exception('--error: gnR1R2 function failed to genrate r1 in 1000 iterations')
    
    r2=np.floor(np.random.uniform(size=(NP0))*NP2) + 1
    for i in range(1,10000):
        pos=np.logical_or((r2 == r1), (r2 == r0))
        if np.sum(pos) == 0:
            break
        else:
            r2[pos]=np.floor(np.random.uniform(size=(sum(pos))) * NP2) + 1
        if i > 1000:
            raise Exception('--error: gnR1R2 function failed to genrate r1 in 1000 iterations')
    
    
    return r1.astype(int), r2.astype(int)

def randFCR(NP,CRm,CRsigma,Fm,Fsigma,*args,**kwargs):

    # this function generate CR according to a normal distribution with mean "CRm" and sigma "CRsigma"
    #If CR > 1, set CR = 1. If CR < 0, set CR = 0.
    #this function generate F  according to a cauchy distribution with location parameter "Fm" and scale parameter "Fsigma"
    #If F > 1, set F = 1. If F <= 0, regenrate F.
    
    ## generate CR
    CR=CRm + CRsigma * np.random.randn(NP)
    CR=np.clip(CR,0,1)
    
    ## generate F
    F=randCauchy(NP,1,Fm,Fsigma)
    F=np.minimum(1,F)
    # we don't want F = 0. So, if F<=0, we regenerate F (instead of trucating it to 0)
    pos=np.argwhere(F <= 0).flatten()
    while pos.size > 0:
        F[pos]=randCauchy(len(pos),1,Fm,Fsigma)
        F=np.minimum(1,F)
        pos=np.argwhere(F <= 0).flatten()
    
    return F, CR
    
    # Cauchy distribution: cauchypdf = @(x, mu, delta) 1/pi*delta./((x-mu).^2+delta^2)
def randCauchy(m,n,mu,delta,*args,**kwargs):

    # http://en.wikipedia.org/wiki/Cauchy_distribution
    result=mu + delta * np.tan(np.pi * (np.random.uniform(size=(m)) - 0.5))
    return result

class ARCHIVE(object):
    """
    Archive class for Ensemble DE algorathim
    
    """
    def __init__(self, NP, dim):
        
        self.NP= NP   #the maximum size of the archive
        self.pop= np.empty((0,dim), dtype=float)
        self.funvalues= np.empty((0), dtype=float)
    
    def UpdateArchive(self, popi, funvalue):
        
        self.pop = np.append(self.pop, popi, axis=0)
        self.funvalues = np.append(self.funvalues, funvalue, axis=0)

        #remove duplicates
        self.pop,self.ind=np.unique(self.pop, axis=0, return_index=True)
        self.funvalues = self.funvalues[self.ind]
        
        #check the archive capacity
        if self.pop.shape[0] > self.NP:
            randpos=np.random.permutation(self.pop.shape[0])[:self.NP]  #keep random NP individuals
            self.pop=self.pop[randpos,:]
            self.funvalues=self.funvalues[randpos]


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

import numpy as np
import pandas as pd
import os
import random
import math
from itertools import repeat
import itertools
import sys, copy, shutil
import subprocess
from multiprocessing.dummy import Pool
from collections import defaultdict
import copy

import random
import matplotlib.pyplot as plt

try: 
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

class ESTUNE:
    """
    A class to parse neorl input template and construct cases for evolution strategy (ES) hyperparameter optimisation

    inputs: 
    The template input file
    Class object from PARSER.py, featuring user input for TUNE
    neorl logo
    """

    def __init__(self, tuneclass, inputfile, tuneblock, logo):
        self.logo=logo
        self.inputfile=inputfile
        self.tuneblock=tuneblock
        self.n_last_episodes=int(self.tuneblock["n_last_episodes"])
        self.ncores=int(self.tuneblock["ncores"])
        self.ncases=int(self.tuneblock["ncases"])

        #---------------------------------------
        # define genetic algorithm parameters
        #---------------------------------------
        self.popsize=10
        if self.ncases < self.popsize:
            self.ngens=1
        else:
            self.ngens=int(self.ncases/self.popsize)
        self.MU=5
        if tuneclass == 'gatune': # ES/GA tune
            print("Performing semi-GA Tune")
            self.INDPB=0.1
        elif tuneclass == 'estune': # ES tune
            print("Performing ES Tune")
            self.INDPB=1.0
        else: # default setting is ES tune
            print("Performing ES Tune")
            self.INDPB=1.0
        self.CXPB=0.5
        self.MUTPB=0.2
        self.ETA=0.6
        self.SMAX=0.5
        self.paramvals=dict()
        self.paraminds=dict()
        self.datatypes=[]

        #-------------------------------
        # construct results directory
        #-------------------------------
        if os.path.exists('./tunecases/'):
            shutil.rmtree('./tunecases/')
            os.makedirs('./tunecases/', exist_ok=True)
        else:
            os.makedirs('./tunecases/', exist_ok=True)
        self.csvlogger='tune.csv'
        self.tunesummary='tunesummary.txt'

        #---------------------------------
        # parse the input template
        #---------------------------------
        with open (self.inputfile, 'r') as input_file_text:
            self.template=input_file_text.readlines()
        
        first=0; last=0
        for i in range(len(self.template)):
            if ('READ TUNE' in self.template[i]):
                first=i
            if ('END TUNE' in self.template[i]):
                last=i
        if first == 0 and last ==0:
            raise ('TUNE card cannot be found')

        del self.template[first: last+1]
        self.template="".join(self.template)

    def tune_count(self):
        
        """
        1- This function uses self.tuneblock, parse it, infer all parameters to be tuned and thier distribution
        2- This function creates GA engine and instantiates the initial population for evolution algorithm
        """
        
        self.param_dict={}
        for item in self.tuneblock:
            if '{' in item and '}' in item and item[0] != '#':
                #-----------------------------------------------------
                # check the existence of the name in the template
                #-----------------------------------------------------
                if item not in self.template:
                    raise ValueError('parameter {} in TUNE block cannot be found in any other block, e.g. DQN, GA, PPO, etc.'.format(item)) 

                item_lst=self.tuneblock[item].split(",")
                item_lst=[item.strip() for item in item_lst] # get rid of white spaces in the splitted values
                #-------------------------------------------------------
                # check if a uniform distribution of floats is identified
                #-------------------------------------------------------
                try:
                    if "float" in item_lst:
                        item_lst[0]=float(item_lst[0])
                        item_lst[1]=float(item_lst[1])
                        self.datatypes.append("float")
                        print ('-- debug: parameter {} has uniform distribution of type --float-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                    elif "u" in item_lst: 
                        item_lst[0]=float(item_lst[0])
                        item_lst[1]=float(item_lst[1])
                        self.datatypes.append("float")
                        print ('-- debug: parameter {} has uniform distribution of type --float-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                except:
                    raise Exception ('--error: TUNE cannot construct the user-given uniform distribution of --floats-- for {} according to (low, high, u) syntax'.format(item))
               
                #---------------------------------------------------
                # check if a random integer distribution is identified
                #---------------------------------------------------
                try:
                    if "int" in item_lst:
                        item_lst[0]=int(item_lst[0])
                        item_lst[1]=int(item_lst[1])
                        self.datatypes.append("int")
                        print ('-- debug: parameter {} has uniform distribution of type --int-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                    elif "randint" in item_lst:
                        item_lst[0]=int(item_lst[0])
                        item_lst[1]=int(item_lst[1])
                        self.datatypes.append("int")
                        print ('-- debug: parameter {} has uniform distribution of type --int-- between {} and {}'.format(item,item_lst[0],item_lst[1]))
                except:
                    raise Exception ('--error: TUNE cannot construct the user-given uniform distribution of --int-- for {} according to (low, high, u) syntax'.format(item))
              
               #-----------------------------------------------------
               # check if a grid is identified
               #-----------------------------------------------------
                try:
                    if "grid" in item_lst:
                        element_lst=[]
                        for element in item_lst:
                            # check if it is an integer
                            not_int=0
                            try:
                                element_lst.append(int(element.strip()))
                            except Exception:
                                not_int=1
                            
                            # else check if the elment is float
                            if not_int:
                                try:
                                    element_lst.append(float(element.strip()))
                                # else consider it a string
                                except Exception:
                                    element_lst.append(str(element.strip()))
                                    
                        item_lst=element_lst
                        self.datatypes.append("grid")
                        print ('-- debug: parameter {} has grid type with values {}'.format(item,item_lst))
                except:
                    raise Exception ('--error: TUNE cannot construct the user-given grid for {} according to the comma-seperated syntax'.format(item))

                self.param_dict[item]=item_lst # Save the final parsed list for parameter {XXX}            
        
        #-----------------------------------------------------
        # infer the bounds for strategy vector 
        #-----------------------------------------------------
        if len(self.param_dict.keys()) <= 10:
            self.SMIN=0.1
        else:
            self.SMIN=1/(len(self.param_dict.keys()))

    def gen_cases(self, x=0):
        """
        This function infers neorl.py path
        """
        self.tune_count()
        self.param_names=list(self.param_dict.keys())
        #-----------------------        
        # Infer neorl.py path
        #-----------------------
        # Find neorl path
        #self.here=os.path.dirname(os.path.abspath(__file__))
        #self.neorl_path=self.here.replace('src/tune','neorl.py') #try to infer neorl.py internally to call neorl inside or neorl
        #self.python_path=self.here.replace('neorl/src/tune','anaconda3/bin/python3') #try to infer python3 path to call neorl inside or neorl

        self.neorl_path=sys.argv[0]
        self.python_path=sys.executable
        print('--debug: NEORLPATH=', self.neorl_path)
        print('--debug: PYTHONPATH=', self.python_path)
        
    def GenES(self):
        """
        Individual generator:
        1- This function uses self.param_dict to obtain bounds for individual parameters
        Returns:
            -ind (list): an individual vector with values samples from inferred distribution 
            -strategy (list): the strategy vector with values between smin and smax     
        """   
        size=len(self.param_dict.keys()) # size of individual
        content=[]
        self.LOW=[] # Lower bounds for the parameters to be tuned
        self.UP=[] # Upper bounds for parameters to be tuned
        for key in list(self.param_dict.keys()):
            if 'int' in self.param_dict[key]:
                content.append(random.randint(self.param_dict[key][0], self.param_dict[key][1]))
            elif 'randint' in self.param_dict[key]:
                content.append(random.randint(self.param_dict[key][0], self.param_dict[key][1]))
            elif 'float' in self.param_dict[key]:
                content.append(random.uniform(self.param_dict[key][0], self.param_dict[key][1]))
            elif 'u' in self.param_dict[key]:
                content.append(random.uniform(self.param_dict[key][0], self.param_dict[key][1]))
            elif 'grid' in self.param_dict[key]:
                self.real_grid=list(self.param_dict[key])
                self.real_grid.remove('grid') # get rid of the 'grid' to avoid sampling it
                self.paramvals[key]=self.real_grid
                content.append(random.sample(self.real_grid, 1)[0])
                self.paraminds[len(content)-1]=key
            else:
                raise Exception('unknown data type is given, either int/randint, float/u, or grid are allowed for parameter distribution types')
            self.LOW.append(self.param_dict[key][0])
            self.UP.append(self.param_dict[key][1])
        ind=list(content)
        size = len(list(self.param_dict.keys()))
        strategy= [random.uniform(self.SMIN, self.SMAX) for _ in range(size)]
        return ind, strategy
    
    def init_pop(self):
        """
        Population initializer
        Returns:
            -pop (dict): initial population in a dictionary form  
        """
        # initialize the population and strategy and run them in parallel (these samples will be used to initialize the memory)
        pop=defaultdict(list)
        
        for i in range(self.popsize):
            #caseid='es_gen{}_ind{}'.format(0,i+1)
            data=self.GenES()
            pop[i].append(data[0])
            pop[i].append(data[1])
        
        if self.ncores > 1: # evaluate warmup in parallel
            core_list=[]
            for key in pop:
                caseid='ind{}'.format(key+1)
                core_list.append([pop[key][0], caseid])
            p=Pool(self.ncores)
            fitness=p.map(self.gen_object, core_list)
            p.close(); p.join()

            [pop[ind].append(fitness[ind]) for ind in range(len(pop))]
        
        else: # evaluate warmup in series
            for key in pop:
                caseid='ind{}'.format(key+1)
                fitness=self.fit(pop[key][0], caseid)
                pop[key].append(fitness)
        return pop # return final pop dictionary with ind, strategy, and fitness

    def fit(self, ind, caseid):
        """
        This function evaluates an individual's fitness
        Inputs:
            -ind (list): an individual whose fitness to evaluate
            -caseid (str): a string that specifies the given individual
        Returns: 
            -mean_reward (float): fitness value 
        """
        try:
            #---------------------------------------------
            # Prepares directories and files for one case
            # --------------------------------------------
            self.param_names=list(self.param_dict.keys())
            i = caseid[3:]

            os.makedirs('./tunecases/case{}'.format(i), exist_ok=True)
            self.new_template=copy.deepcopy(self.template)
            for j in range (len(self.param_names)):
                self.new_template=self.new_template.replace(str(self.param_names[j]), str(ind[j]))
            
            filename='./tunecases/case{}/case{}.inp'.format(i, i)
            with open (filename, 'w') as fout:
                fout.writelines(self.new_template)
                
            # copy external files into the new directory, if extfiles card exists
            if 'extfiles' in self.tuneblock.keys():
                if self.tuneblock['extfiles']:
                    print('--debug: external files are identified, copying them into each case directory')
                    for item in self.tuneblock['extfiles']:
                        os.system('cp -r {} ./tunecases/case{}/'.format(item, i))

            casenum = caseid[3:]
            print('--------------------------------------------------')
            print('Running TUNE Case {}/{}: {}'.format(casenum, self.ncases, ind))
            subprocess.call([self.python_path, self.neorl_path, '-i', 'case{}.inp'.format(casenum)], cwd='./tunecases/case{}/'.format(casenum))  # this exceutes neorl for this case.inp
            print('--------------------------------------------------')
            
            #--------------------------------------------------------------------------------------------------------------
            # Try to infer the _out.csv file in the directory since only one method is allowed
            csvfile=[f for f in os.listdir('./tunecases/case{}/case{}_log/'.format(casenum, casenum)) if f.endswith('_out.csv')]
            if len(csvfile) > 1:
                raise Exception ('multiple *_out.csv files can be found in the logger of TUNE, only one is allowed')
            #--------------------------------------------------------------------------------------------------------------
            reward_lst=pd.read_csv('./tunecases/case{}/case{}_log/{}'.format(casenum,casenum, csvfile[0]), usecols=['reward']).values
            mean_reward=np.mean(reward_lst[-self.n_last_episodes:])
            max_reward=np.max(reward_lst)
            
            with open (self.csvlogger, 'a') as fout:
                fout.write(str(casenum) +',')
                [fout.write(str(item) + ',') for item in ind]
                fout.write(str(mean_reward) + ',' + str(max_reward) + '\n')
                
            return mean_reward
        
        except:
            print('--error: case{}.inp failed during execution'.format(casenum))
            
            return 'case{}.inp:failed'.format(casenum)
            
    def gen_object(self, inp):
        """
        This is a worker for the multiprocess Pool 
        Inputs:
            -inp (list of lists): contains data for each core [[ind1, caseid1], ...,  [indN, caseidN]]
        Returns: 
            -fitness value (float)
        """
        return self.fit(inp[0], inp[1])
    
    def select(self, pop):
        """
        Selection function sorts the population from max to min based on fitness and selects the k best
        Inputs:
            -pop (dict): population in dictionary structure
            -k (int): top k individuals are selected
        Returns:
            -best_dict (dict): the new orded dictionary with top k selected
        """
        k=self.MU
        pop=list(pop.items())
        pop.sort(key=lambda e: e[1][2], reverse=True)
        sorted_dict=dict(pop[:k])

        # This block creates a new dict where keys are reset to 0 ... k in order to avoid unordered keys after sort 
        best_dict=defaultdict(list)
        index=0
        for key in sorted_dict:
            best_dict[index].append(sorted_dict[key][0])
            best_dict[index].append(sorted_dict[key][1])
            best_dict[index].append(sorted_dict[key][2])
            index+=1 

        sorted_dict.clear()
        return best_dict
    
    def cx(self, ind1, ind2, strat1, strat2):
        """
        Executes a classical two points crossover on both the individuals and their strategy. 
        The individuals/strategies should be a list. The crossover points for the individual and the 
        strategy are the same. 

        Inputs:
            -ind1 (list): The first individual participating in the crossover. 
            -ind2 (list): The second individual participating in the crossover.
            -strat1 (list): The first evolution strategy participating in the crossover. 
            -strat2 (list): The second evolution strategy 
        Returns:
            - The new ind1, ind2, strat1, strat2, after crossover in list form
        """
        
        #for item in ind1:
        #    print('individual 1', type(item))
        #for item in ind2:
        #    print('individual 2', type(item))
        #for item in strat1:
        #    print('strategy 1', type(item))
        #for item in strat2:
        #    print('strategy 2', type(item))
        
        size = min(len(ind1), len(ind2))

        pt1 = random.randint(1, size)
        pt2 = random.randint(1, size-1)
        if pt2 >= pt1:
            pt2 +=1
        else:
            pt1, pt2 = pt2, pt1
        
        ind1[pt1:pt2], ind2[pt1:pt2] = ind2[pt1:pt2], ind1[pt1:pt2]
        strat1[pt1:pt2], strat2[pt1:pt2] = strat2[pt1:pt2], strat1[pt1:pt2]

        return ind1, ind2, strat1, strat2 
    
    def mutES(self, ind, strat):
        """
        Mutate an evolution strategy according to mixed Discrete/Continuous mutation rules 
        Input:
            -ind (list): individual to be mutated
            -strat (list): individual strategy to be mutated 
        Returns:
            -ind (list): new individual after mutation
            -strat (list): individual strategy after mutation
        """
        size=len(ind)
        tau=1/np.sqrt(2*size)
        tau_prime=1/np.sqrt(2*np.sqrt(size))
        
        for i in range(size):
            # Grid distribution received
            if self.datatypes[i] == "grid":
            #if i in self.paraminds.keys():
                norm=random.gauss(0,1)
                # modify the ind strategy
                strat[i] = 1/(1+(1-strat[i])/strat[i]*np.exp(-tau*norm-tau_prime*random.gauss(0,1)))
                # make a transformation of strategy to ensure it is between smin, smax 
                y=(strat[i]-self.SMIN)/(self.SMAX-self.SMIN)
                if np.floor(y) % 2 == 0:
                    y_prime=np.abs(y-np.floor(y))
                else:
                    y_prime=1-np.abs(y-np.floor(y))
                strat[i] = self.SMIN + (self.SMAX-self.SMIN)*y_prime

                # check if this attribute is mutated based on the updated strategy
                if random.random() < strat[i]:
                    # make a list of possibilities after excluding the current value to enforce mutation
                    paramname=self.paraminds[i]
                    ind[i]=random.sample(self.paramvals[paramname], 1)[0]

            # Random integer distribution received
            elif self.datatypes[i] == "int":
                norm=random.gauss(0,1)
                # modify the ind strategy 
                strat[i] = 1/(1+(1-strat[i])/strat[i]*np.exp(-tau*norm-tau_prime*random.gauss(0,1)))
                # make a transformation of strategy to ensure it is between smin, smax 
                y=(strat[i]-self.SMIN)/(self.SMAX-self.SMIN)
                if np.floor(y) % 2 == 0:
                    y_prime=np.abs(y-np.floor(y))
                else:
                    y_prime=1-np.abs(y-np.floor(y))
                strat[i] = self.SMIN + (self.SMAX-self.SMIN)*y_prime

                # check if this attribute is mutated based on the updated strategy 
                #if random.random() < strat[i]:
                # make a list of possibilities after excluding the current value to enforce mutation
                choices=list(range(self.LOW[i], self.UP[i]+1))
                choices.remove(ind[i])
                ind[i] = random.choice(choices)

            # Uniform float distribution received
            elif self.datatypes[i] == "float":
                norm=random.gauss(0,1)
                if random.random() < self.INDPB: # this indicates whether ind/strategy to be mutated or not for this float variable
                    strat[i] *= np.exp(tau*norm + tau_prime * random.gauss(0,1)) # normal mutation strategy
                    ind[i] += strat[i] * random.gauss(0,1) # update the individual position
                
                # check if the new individual falls within lower/uppder boundaries
                if ind[i] < self.LOW[i]:
                    ind[i] = self.LOW[i]
                if ind[i] > self.UP[i]:
                    ind[i] = self.UP[i]
                
            else:
                raise Exception('ES mutation strategy works with int, float, or grid distributions, the type provided cannot be interpreted')
        
        return ind, strat
    
    def GenOffspring(self, pop):
        """
        This function generates the offspring by applying crossover, mutation, OR reproduction. 
        Inputs:
            -pop (dict): population in dictionary structure
        Returns:
            -offspring (dict): new modified population in dictionary structure
        """

        pop_indices=list(range(0,len(pop)))
        offspring=defaultdict(list)
        for i in range(self.popsize):
            alpha=random.random()
            #----------------------
            # Crossover
            #----------------------
            if alpha < self.CXPB:
                index1, index2=random.sample(pop_indices,2)
                ind1, ind2, strat1, strat2=self.cx(ind1=list(pop[index1][0]), ind2=list(pop[index2][0]),
                                                        strat1=list(pop[index1][1]), strat2=list(pop[index2][1]))
                offspring[i].append(ind1)
                offspring[i].append(strat1)
                #print('crossover is done for sample {} between {} and {}'.format(i,index1,index2))
            #----------------------
            # Mutation
            #----------------------
            elif alpha < self.CXPB + self.MUTPB:  # Apply mutation
                index = random.choice(pop_indices)
                
                ind, strat=self.mutES(ind=list(pop[index][0]), strat=list(pop[index][1]))
                offspring[i].append(ind)
                offspring[i].append(strat)
                #print('mutation is done for sample {} based on {}'.format(i,index))
            #------------------------------
            # Reproduction from population
            #------------------------------
            else:
                index=random.choice(pop_indices)
                offspring[i].append(pop[index][0])
                offspring[i].append(pop[index][1])
                #print('reproduction is done for sample {} based on {}'.format(i,index))
        return offspring 

    def run_cases(self):
        """
        This function runs the evolutioanry algorithm over self.ngens generations. 
        """
        #------------------------------
        # Begin the evolution process
        #------------------------------
        with open (self.csvlogger, 'w') as fout:
            fout.write('caseid, ')
            [fout.write(item + ',') for item in self.param_names]
            fout.write('mean_reward,max_reward\n')

        #print('PARAM dict', self.param_dict)
        #print('PARAM types', self.datatypes)
        self.population=self.init_pop()
        case_idx=0
        self.currentcase=self.popsize+1
        for gen in range(1, self.ngens): 
            case_idx=0
            caseids=['ind{}'.format(ind) for ind in range(self.currentcase, self.currentcase+self.popsize+1)]
            # Vary the population and generate new offspring
            offspring=self.GenOffspring(pop=self.population)

            # Evaluate the individuals with invalid fitness using multiprocessing Pool
            if self.ncores > 1:
                core_list=[]
                for key in offspring:
                    core_list.append([offspring[key][0], caseids[case_idx]])
                    case_idx+=1
                # initialize a pool
                p=Pool(self.ncores)
                fitness=p.map(self.gen_object, core_list)
                p.close(); p.join()

                [offspring[ind].append(fitness[ind]) for ind in range(len(offspring))]
            else:
                for ind in range(len(offspring)):
                    fitness=self.fit(offspring[ind][0], caseids[case_idx])
                    case_idx+=1
                    offspring[ind].append(fitness)
            
            self.currentcase+=self.popsize
            # Select the next generation population 
            self.population = copy.deepcopy(self.select(pop=offspring))
        

        csvdata=pd.read_csv('tune.csv')
        asc_data=csvdata.sort_values(by=['caseid'],ascending=True)
        des_data=csvdata.sort_values(by=['mean_reward'],ascending=False)
        des_data2=csvdata.sort_values(by=['max_reward'],ascending=False)
        asc_data.to_csv('tune.csv', index=False)

        mean = np.mean(des_data.iloc[:,4:5])
        totalmean=mean.tolist()[0]
        
        try:
            failed_cases=len([print ('failed') for item in self.population if isinstance(item, str)])
        except:
            failed_cases='NA'
        
        print ('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print('Mean Rewards for all cases=', totalmean)
        print ('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print ('All TUNE CASES ARE COMPLETED')
        print ('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
        print('--debug: Check tunesummary.txt file for best hyperparameters found')
        print('--debug: Check tune.csv file for complete csv logger of all cases results')
        print('--debug: Check tunecases directory for case-by-case detailed results')
        
        with open ('tunesummary.txt', 'w') as fout:
            
            #fout.write(self.logo)
            fout.write('*****************************************************\n')
            fout.write('Summary for the TUNE case \n')
            fout.write('*****************************************************\n')
            fout.write('Number of cases evaluated: {} \n'.format(self.ncases))
            fout.write('Number of failed cases: {} \n'.format(failed_cases))
            fout.write('Parameter names: {} \n'.format(self.param_names))
            fout.write('Parameter values: {} \n '.format(self.param_dict))
            fout.write ('--------------------------------------------------------------------------------------\n')
            
            if des_data.shape[0] < 20:
                top=des_data.shape[0]
                fout.write ('Top {} hyperparameter configurations ranked according to MEAN reward \n'.format(top))
                fout.write(des_data.iloc[:top].to_string(index=False))
            else:
                top=20
                fout.write ('Top {} hyperparameter configurations ranked according to MEAN reward \n'.format(top))
                fout.write(des_data.iloc[:top].to_string(index=False))
            fout.write ('\n')
            fout.write ('--------------------------------------------------------------------------------------\n')
            if des_data2.shape[0] < 20:
                top=des_data2.shape[0]
                fout.write ('Top {} hyperparameter configurations ranked according to MAX reward \n'.format(top))
                fout.write(des_data2.iloc[:top].to_string(index=False))
            else:
                top=20
                fout.write ('Top {} hyperparameter configurations ranked according to MAX reward \n'.format(top))
                fout.write(des_data2.iloc[:top].to_string(index=False))


# flake8: noqa F403
from neorl.rl.baselines.shared.console_util import fmt_row, fmt_item, colorize
from neorl.rl.baselines.shared.dataset import Dataset
from neorl.rl.baselines.shared.math_util import discount, discount_with_boundaries, explained_variance, \
    explained_variance_2d, flatten_arrays, unflatten_vector
from neorl.rl.baselines.shared.misc_util import zipsame, set_global_seeds, boolean_flag
from neorl.rl.baselines.shared.base_class import BaseRLModel, ActorCriticRLModel, OffPolicyRLModel, SetVerbosity, \
    TensorboardWriter
from neorl.rl.baselines.shared.cmd_util import make_vec_env



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu Jul  2 16:25:37 2020
#
#@author: Majdi
#"""

import random
from math import sin, cos, pi, exp, e, sqrt
from operator import mul
from functools import reduce

def sphere(individual):
    """Sphere test objective function.
    """
    return sum(x**2 for x in individual)

def cigar(individual):
    """Cigar test objective function.
    """
    return individual[0]**2 + 1e6 * sum(gene * gene for gene in individual)

def rosenbrock(individual):  
    """Rosenbrock test objective function.
    """
    return sum(100 * (x * x - y)**2 + (1. - x)**2 \
                   for x, y in zip(individual[:-1], individual[1:]))

def ackley(individual):
    """Ackley test objective function.
    """
    N = len(individual)
    return 20 - 20 * exp(-0.2*sqrt(1.0/N * sum(x**2 for x in individual))) \
            + e - exp(1.0/N * sum(cos(2*pi*x) for x in individual))
            
def bohachevsky(individual):
    """Bohachevsky test objective function.
    """
    return sum(x**2 + 2*x1**2 - 0.3*cos(3*pi*x) - 0.4*cos(4*pi*x1) + 0.7 
                for x, x1 in zip(individual[:-1], individual[1:]))

def griewank(individual):
    """Griewank test objective function.
    """
    return 1.0/4000.0 * sum(x**2 for x in individual) - \
        reduce(mul, (cos(x/sqrt(i+1.0)) for i, x in enumerate(individual)), 1) + 1
            
def rastrigin(individual):
    """Rastrigin test objective function.
    """     
    return 10 * len(individual) + sum(gene * gene - 10 * \
                        cos(2 * pi * gene) for gene in individual)

def rastrigin_scaled(individual):
    """Scaled Rastrigin test objective function.
    """
    N = len(individual)
    return 10*N + sum((10**(i/(N-1))*x)**2 - 
                      10*cos(2*pi*10**(i/(N-1))*x) for i, x in enumerate(individual))

def rastrigin_skew(individual):
    """Skewed Rastrigin test objective function.
    """
    N = len(individual)
    return 10*N + sum((10*x if x > 0 else x)**2 
                    - 10*cos(2*pi*(10*x if x > 0 else x)) for x in individual)
    

def schaffer(individual):
    """Schaffer test objective function.
    """
    return sum((x**2+x1**2)**0.25 * ((sin(50*(x**2+x1**2)**0.1))**2+1.0) 
                for x, x1 in zip(individual[:-1], individual[1:]))

def schwefel(individual):
    """Schwefel test objective function.
    """    
    N = len(individual)
    return 418.9828872724339*N-sum(x*sin(sqrt(abs(x))) for x in individual)

def alpinen1(individual): 
    return sum(abs(x * sin(x) + 0.1 * x) for x in individual)

def alpinen2(individual):
    return reduce(mul, (sqrt(x)*sin(x) for x in individual), 1)

def brown(individual):
    scores=0
    x=[item**2 for item in individual]
    for i in range(len(individual) - 1):
        scores += x[i] ** (x[i+1] + 1) + x[i+1]**(x[i] + 1)    
    return scores

def expo(individual):
    return -exp(-0.5*sum(x**2 for x in individual))

def yang(individual):
    return sum(random.random()*abs(x)**i for i, x in enumerate(individual,start=1))

def yang2(individual):
    return sum(abs(x) for x in individual) * exp(-sum(sin(x**2) for x in individual))

def yang3(individual):
    beta=15
    m=5
    return exp(-sum((x/beta)**(2*m) for x in individual))  \
            - 2*exp(-sum(x**2 for x in individual)) * reduce(mul, (cos(x)**2 for x in individual), 1)

def yang4(individual):
    return (sum(sin(x)**2 for x in individual) - exp(-sum(x**2 for x in individual))) \
            * exp(-sum(sin(sqrt(abs(x)))**2 for x in individual))
            
def zakharov (individual):
    return sum(x**2 for x in individual) + sum(0.5*i*x for i, x in enumerate(individual,start=1))**2 \
            + sum(0.5*i*x for i, x in enumerate(individual,start=1))**4
            
def salomon (individual):
    return 1-cos(2*pi*sqrt(sum(x**2 for x in individual))) + 0.1*sqrt(sum(x**2 for x in individual))

def st(individual):
    return 0.5*sum(x**4 - 16*x**2 + 5*x for x in individual)

def shubert(individual):  #2d function, global minima at -186.7309
    score=1
    for i in range(len(individual)):
        score*= sum(x*cos((x+1)*individual[i] + x) for x in [1,2,3,4,5])
    return score

def ridge(individual): 
    d=1
    alpha=0.5
    return individual[0] + d*sum(individual[i]**2 for i in range(1,len(individual)))**alpha

def powell(individual):
    return sum(abs(x)**(i+1) for i,x in enumerate (individual,start=1))

#def periodic(individual):
#    return 1+sum(sin(x)**2 for x in individual)-0.1*exp(sum(x**2 for x in individual))

def qing(individual):
    return sum((x**2 - i)**2 for i,x in enumerate (individual,start=1))

def quartic(individual):
    return sum(i*x**4 for i,x in enumerate (individual,start=1)) + random.random()

def happycat(individual):
    l2=sum(x**2 for x in individual)
    n=len(individual)
    alpha=1/8
    return ((l2-n)**2)**alpha + 1/n*(0.5*l2 + sum(individual)) + 0.5

def schwefel2(individual):
    return sum(abs(x) for x in individual) + reduce(mul, (abs(x) for x in individual), 1)

def dixonprice(individual):
    score=0
    for i in range(1,len(individual)):
         score+= (i+1)*(2*individual[i]**2-individual[i-1])**2
    score += (individual[0]-1)**2
    return  score

def levy(individual):

    d = len(individual)
    
    w=[1 + (x - 1)/4 for x in individual]
    
    term1 = sin(pi*w[0])**2
    term3 = (w[-1]-1)**2 * (1+sin(2*pi*w[-1])**2)
    
    term2 = 0;
    for i in range (d-1):
        term2 += (w[i]-1)**2 * (1+10*sin(pi*w[i]+1)**2)
        
    return term1 + term2 + term3

all_functions = [
    sphere,
    cigar,
    rosenbrock,
    bohachevsky,
    griewank,
    rastrigin,
    ackley,
    rastrigin_scaled,
    rastrigin_skew,
    schaffer,
    schwefel,
    schwefel2,
#    alpinen1,
#    alpinen2,
    brown,
    expo,
    yang,
    yang2,
    yang3,
    yang4,
    zakharov,
    salomon,
    st,
    shubert,
#    ridge,
    powell,
    qing,
    quartic,
    happycat,
    dixonprice,
    levy
]

all_names = [
    'sphere',
    'cigar',
    'rosenbrock',
    'bohachevsky',
    'griewank',
    'rastrigin',
    'ackley',
    'rastrigin_scaled',
    'rastrigin_skew',
    'schaffer',
    'schwefel',
    'schwefel2',
#    alpinen1,
#    alpinen2,
    'brown',
    'expo',
    'yang',
    'yang2',
    'yang3',
    'yang4',
    'zakharov',
    'salomon',
    'st',
    'shubert',
#    'ridge',
    'powell',
    'qing',
    'quartic',
    'happycat',
    'dixonprice',
    'levy'
]

all_bounds = [
    [-5.12, 5.12],#sphere
    [-100, 100],#cigar
    [-5, 10],#rosenbrock
    [-100, 100],#bohachevsky
    [-600, 600],#griewank
    [-5.12, 5.12],#rastrigin
    [-32.768, 32.768],#ackley
    [-5.12, 5.12],#rastrigin_scaled
    [-5.12, 5.12],#rastrigin_skew
    [-100, 100],#schaffer
    [-500, 500],#schwefel
    [-100, 100],#schwefel2
    #alpinen1
    #alpinen2
    [-1, 4],#brown
    [-1, 1],#expo
    [-5, 5],#yang
    [-2*pi, 2*pi],#yang2
    [-20, 20],#yang3
    [-10, 10],#yang4
    [-5, 10],#zakharov
    [-100, 100],#salomon
    [-5, 5],#st
    [-10, 10],#shubert
    #ridge
    [-1,1],#powell
    [-500,500],#qing
    [-1.28, 1.28],#quartic
    [-2, 2],#happycat
    [-10, 10],#dixonprice
    [-10, 10]#levy
]

all_ndim = [
    'arbitrary',#sphere
    'arbitrary',#cigar
    'arbitrary',#rosenbrock
    'arbitrary',#bohachevsky
    'arbitrary',#griewank
    'arbitrary',#rastrigin
    'arbitrary',#ackley
    '2+',#rastrigin_scaled
    'arbitrary',#rastrigin_skew
    'arbitrary',#schaffer
    'arbitrary',#schwefel
    'arbitrary',#schwefel2
    #alpinen1
    #alpinen2
    '2+',#brown
    'arbitrary',#expo
    'arbitrary',#yang
    'arbitrary',#yang2
    'arbitrary',#yang3
    'arbitrary',#yang4
    'arbitrary',#zakharov
    'arbitrary',#salomon
    'arbitrary',#st
    '2',#shubert
    #ridge
    'arbitrary',#powell
    'arbitrary',#qing
    'arbitrary',#quartic
    'arbitrary',#happycat
    'arbitrary',#dixonprice
    'arbitrary'#levy
]

all_minima_x = [
    0,#sphere
    0,#cigar
    1,#rosenbrock
    0,#bohachevsky
    0,#griewank
    0,#rastrigin
    0,#ackley
    0,#rastrigin_scaled
    0,#rastrigin_skew
    0,#schaffer
    420.96874636,#schwefel
    0,#schwefel2
    #alpinen1
    #alpinen2
    0,#brown
    0,#expo
    0,#yang
    0,#yang2
    0,#yang3
    0,#yang4
    0,#zakharov
    0,#salomon
    -2.903534,#st
    [-7.0825, 4.8580],#shubert, multiple x points for minima
    #ridge
    0,#powell
    "+/-[sqrt(1), sqrt(2), ..., sqrt(ndim)]",#qing
    "-1 or 2",#quartic
    -1,#happycat
    "2^(-(2^i-2)/2^i) for i in [1,2,3,...ndim]",#dixonprice
    1#levy
]



import numpy as np


def conjugate_gradient(f_ax, b_vec, cg_iters=10, callback=None, verbose=False, residual_tol=1e-10):
    """
    conjugate gradient calculation (Ax = b), bases on
    https://epubs.siam.org/doi/book/10.1137/1.9781611971446 Demmel p 312

    :param f_ax: (function) The function describing the Matrix A dot the vector x
                 (x being the input parameter of the function)
    :param b_vec: (numpy float) vector b, where Ax = b
    :param cg_iters: (int) the maximum number of iterations for converging
    :param callback: (function) callback the values of x while converging
    :param verbose: (bool) print extra information
    :param residual_tol: (float) the break point if the residual is below this value
    :return: (numpy float) vector x, where Ax = b
    """
    first_basis_vect = b_vec.copy()  # the first basis vector
    residual = b_vec.copy()  # the residual
    x_var = np.zeros_like(b_vec)  # vector x, where Ax = b
    residual_dot_residual = residual.dot(residual)  # L2 norm of the residual

    fmt_str = "%10i %10.3g %10.3g"
    title_str = "%10s %10s %10s"
    if verbose:
        print(title_str % ("iter", "residual norm", "soln norm"))

    for i in range(cg_iters):
        if callback is not None:
            callback(x_var)
        if verbose:
            print(fmt_str % (i, residual_dot_residual, np.linalg.norm(x_var)))
        z_var = f_ax(first_basis_vect)
        v_var = residual_dot_residual / first_basis_vect.dot(z_var)
        x_var += v_var * first_basis_vect
        residual -= v_var * z_var
        new_residual_dot_residual = residual.dot(residual)
        mu_val = new_residual_dot_residual / residual_dot_residual
        first_basis_vect = residual + mu_val * first_basis_vect

        residual_dot_residual = new_residual_dot_residual
        if residual_dot_residual < residual_tol:
            break

    if callback is not None:
        callback(x_var)
    if verbose:
        print(fmt_str % (i + 1, residual_dot_residual, np.linalg.norm(x_var)))
    return x_var



from neorl import PPO2, A2C, ACKTR, ACER, DQN
from neorl import MlpPolicy
from neorl import DQNPolicy
from neorl import RLLogger
from neorl import CreateEnvironment
import math

def test_mixedrl():
    
    def Vessel(individual):
        """
        Pressure vesssel design
        x1: thickness (d1)  --> discrete value multiple of 0.0625 in 
        x2: thickness of the heads (d2) ---> categorical value from a pre-defined grid
        x3: inner radius (r)  ---> cont. value between [10, 200]
        x4: length (L)  ---> cont. value between [10, 200]
        """
    
        x=individual.copy()
        x[0] *= 0.0625   #convert d1 to "in" 
    
        y = 0.6224*x[0]*x[2]*x[3]+1.7781*x[1]*x[2]**2+3.1661*x[0]**2*x[3]+19.84*x[0]**2*x[2];
    
        g1 = -x[0]+0.0193*x[2];
        g2 = -x[1]+0.00954*x[2];
        g3 = -math.pi*x[2]**2*x[3]-(4/3)*math.pi*x[2]**3 + 1296000;
        g4 = x[3]-240;
        g=[g1,g2,g3,g4]
        
        phi=sum(max(item,0) for item in g)
        eps=1e-5 #tolerance to escape the constraint region
        penality=1e7 #large penality to add if constraints are violated
        
        if phi > eps:  
            fitness=phi+penality
        else:
            fitness=y
        return fitness
    
    for item in ['float', 'int', 'grid', 'float/int', 'float/grid', 'int/grid', 'mixed']:
        bounds = {}
        t=60
        btype=item  #float, int, grid, float/int, float/grid, int/grid, mixed. 
        
        if btype=='mixed':
            bounds['x1'] = ['int', 1, 99]
            bounds['x2'] = ['grid', (0.0625, 0.125, 0.1875, 0.25, 0.3125, 0.375, 0.4375, 0.5, 0.5625, 0.625)]
            bounds['x3'] = ['float', 10, 200]
            bounds['x4'] = ['float', 10, 200]
            bounds['x5'] = ['grid', ('Hi', 'Bye', 'New')]
            bounds['x6'] = ['int', -5, 5]
        
        elif btype=='int/grid':      
            bounds['x1'] = ['int', 1, 20]
            bounds['x2'] = ['grid', (0.0625, 0.125, 0.1875, 0.25, 0.3125, 0.375, 0.4375, 0.5, 0.5625, 0.625)]
            bounds['x3'] = ['int', 10, 200]
            bounds['x4'] = ['int', 10, 200]
            bounds['x5'] = ['grid', ('Hi', 'Bye', 'New')]
            bounds['x6'] = ['int', -5, 5]
        
        elif btype=='float/grid':      
            bounds['x1'] = ['float', 1, 20]
            bounds['x2'] = ['grid', (0.0625, 0.125, 0.1875, 0.25, 0.3125, 0.375, 0.4375, 0.5, 0.5625, 0.625)]
            bounds['x3'] = ['float', 10, 200]
            bounds['x4'] = ['float', 10, 200]
            bounds['x5'] = ['grid', ('Hi', 'Bye', 'New')]
            bounds['x6'] = ['float', -5, 5]
            
        elif btype=='float/int':      
            bounds['x1'] = ['int', 1, 20]
            bounds['x2'] = ['float', 1, 20]
            bounds['x3'] = ['int', 10, 200]
            bounds['x4'] = ['float', 10, 200]
            bounds['x5'] = ['float', -5, 5]
            bounds['x6'] = ['int', -5, 5]
        
        elif btype=='float':      
            bounds['x1'] = ['float', 1, 20]
            bounds['x2'] = ['float', 1, 20]
            bounds['x3'] = ['float', 10, 200]
            bounds['x4'] = ['float', 10, 200]
            bounds['x5'] = ['float', -5, 5]
            bounds['x6'] = ['float', -5, 5]
            
        elif btype=='int':      
            bounds['x1'] = ['int', 1, 20]
            bounds['x2'] = ['int', 1, 20]
            bounds['x3'] = ['int', 10, 200]
            bounds['x4'] = ['int', 10, 200]
            bounds['x5'] = ['int', -5, 5]
            bounds['x6'] = ['int', -5, 5]
            
        elif btype=='grid':      
            bounds['x1'] = ['grid', (0.0625, 0.125, 0.375, 0.4375, 0.5625, 0.625)]
            bounds['x2'] = ['grid', (0.0625, 0.125, 0.1875, 0.25, 0.3125, 0.375, 0.4375, 0.5, 0.5625, 0.625)]
            bounds['x3'] = ['grid', (1,2,3,4,5)]
            bounds['x4'] = ['grid', (32,64,128)]
            bounds['x5'] = ['grid', ('Hi', 'Bye', 'New')]
            bounds['x6'] = ['grid', ('Cat', 'Dog', 'Bird', 'Fish')]
        
        #---------------------------------
        # PPO
        #---------------------------------
        env=CreateEnvironment(method='ppo', fit=Vessel, bounds=bounds, mode='min', episode_length=50)
        cb=RLLogger(check_freq=1)
        ppo = PPO2(policy=MlpPolicy, env=env, n_steps=4, seed=1)
        ppo.learn(total_timesteps=t, callback=cb)
        print('--------------- PPO results ---------------')
        print('The best value of x found:', cb.xbest)
        print('The best value of y found:', cb.rbest)
        
        assert Vessel(cb.xbest) - cb.rbest < 1
        
        #---------------------------------
        # A2C
        #---------------------------------
        cb=RLLogger(check_freq=1)
        a2c = A2C(policy=MlpPolicy, env=env, n_steps=4, seed=1)
        a2c.learn(total_timesteps=t, callback=cb)
        print('--------------- A2C results ---------------')
        print('The best value of x found:', cb.xbest)
        print('The best value of y found:', cb.rbest)
        assert Vessel(cb.xbest) - cb.rbest < 1
        
        #---------------------------------
        # ACKTR
        #---------------------------------
        cb=RLLogger(check_freq=1)
        acktr = ACKTR(policy=MlpPolicy, env=env, n_steps=4)
        acktr.learn(total_timesteps=t, callback=cb)
        print('--------------- ACKTR results ---------------')
        print('The best value of x found:', cb.xbest)
        print('The best value of y found:', cb.rbest)
        assert Vessel(cb.xbest) - cb.rbest < 1
        
        
        if btype in ['int', 'grid', 'int/grid']:
            #---------------------------------
            # ACER
            #---------------------------------
            disc_env=CreateEnvironment(method='acer', fit=Vessel, mode='min', 
                                       bounds=bounds, episode_length=50)
            cb=RLLogger(check_freq=1)
            acer = ACER(MlpPolicy, env=disc_env, n_steps=4)
            acer.learn(total_timesteps=t, callback=cb)
            print('--------------- ACER results ---------------')
            print('The best value of x found:', cb.xbest)
            print('The best value of y found:', cb.rbest)
            assert Vessel(cb.xbest) - cb.rbest < 1
            
            #---------------------------------
            # DQN
            #---------------------------------
            cb=RLLogger(check_freq=1)
            dqn = DQN(DQNPolicy, env=disc_env)
            dqn.learn(total_timesteps=t, callback=cb)
            print('--------------- DQN results ---------------')
            print('The best value of x found:', cb.xbest)
            print('The best value of y found:', cb.rbest)
            assert Vessel(cb.xbest) - cb.rbest < 1
    
    return

test_mixedrl()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Sun Aug 15 2021
#
#@author: Paul
#"""


import random
import numpy as np
import math
import time
import joblib
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual

class CS(object):
    """
    Cuckoo Search Algorithm
    
    :param mode: (str) problem type, either ``min`` for minimization problem or ``max`` for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param ncuckoos: (int) number of cuckoos or nests in the population: one cuckoos per nest. Default value is 15.
    :param pa: (float) a scalar value for the coefficient that controls exploration/exploitation, 
                            i.e. fraction of the cuckoos/nests that will be replaced by the new cuckoos/nests.
    :param int_transform: (str) method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors (must be ``<= ncuckoos``)
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, bounds, fit, ncuckoos=15, pa=0.25, int_transform='nearest_int', ncores=1, seed=None):
        
        set_neorl_seed(seed)
        assert ncores <= ncuckoos, '--error: ncores ({}) must be less than or equal than ncuckoos ({})'.format(ncores, ncuckoos)
        
        self.mode=mode #  mode for optimization: CS only solves a minimization problem.
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
          
        self.int_transform=int_transform
        if int_transform not in ["nearest_int", "sigmoid", "minmax"]:
            raise ValueError('--error: int_transform entered by user is invalid, must be `nearest_int`, `sigmoid`, or `minmax`')
            
        self.bounds=bounds
        self.ncores = ncores
        self.ncuckoos = ncuckoos
        self.pa = pa # Discovery rate of parasitic eggs/solutions
        
        self.var_type = np.array([bounds[item][0] for item in bounds])# infer variable types 
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=np.array([self.bounds[item][1] for item in self.bounds])
        self.ub=np.array([self.bounds[item][2] for item in self.bounds])

    def init_sample(self, bounds):
        #"""
        #Initialize the initial population of cuckoos/nests
        #"""
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        return np.array(indv)

    def eval_cuckoos(self,newnest = None):
        #---------------------
        # Fitness calcs
        #---------------------
        core_lst=[]
        if newnest is None:
            for case in range (0, self.Positions.shape[0]):
                core_lst.append(self.Positions[case, :])
            if self.ncores > 1:
                with joblib.Parallel(n_jobs=self.ncores) as parallel:
                    fitness_lst=parallel(joblib.delayed(self.fit_worker)(item) for item in core_lst)
            else:
                fitness_lst=[]
                for item in core_lst:
                    fitness_lst.append(self.fit_worker(item))
        else:# fitness of new cuckoo versus old cuckoo must also be compared. newnest are the new cuckoos
            for case in range (0, newnest.shape[0]):
                core_lst.append(newnest[case, :])
            if self.ncores > 1:
                with joblib.Parallel(n_jobs=self.ncores) as parallel:
                    fitness_lst=parallel(joblib.delayed(self.fit_worker)(item) for item in core_lst)
            else:
                fitness_lst=[]
                for item in core_lst:
                    fitness_lst.append(self.fit_worker(item))
            
        return fitness_lst

    def select(self, pos, fit):
    
        best_fit=np.min(fit)
        min_idx=np.argmin(fit)
        best_pos=pos[min_idx,:].copy()  
        return best_pos, best_fit 
        
    def ensure_bounds(self, vec):
        vec_new = []
        # cycle through each variable in vector 
        for i, (key, val) in enumerate(self.bounds.items()):
            # variable exceedes the minimum boundary
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
            # variable exceedes the maximum boundary
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
            # the variable is fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
        return vec_new

    def fit_worker(self, x):
        #This worker is for parallel calculations
        # Clip the cuckoo with position outside the lower/upper bounds and return same position
        x=self.ensure_bounds(x)
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)
        # Calculate objective function for each search agent
        fitness = self.fit(x)
        return fitness

    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 
        #handy function to be used within CS phases

        #Params:
        #vec - salp position in vector/list form

        #Return:
        #vec - updated salp position vector with discrete values
        #"""

        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.a,
                                               method=self.int_transform,
                                               )

        return vec

    def UpdateCuckoos(self):
        # perform Levy flights to generate 
        # self.ncuckoos new Cuckoos
        tempnest = np.zeros((self.ncuckoos, self.dim))
        tempnest = np.array(self.Positions).copy()
        beta = 3 / 2
        sigma = (
            math.gamma(1 + beta)
            * math.sin(math.pi * beta / 2)
            / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))
        ) ** (1 / beta)
        s = np.zeros(self.dim)
        for j in range(0, self.ncuckoos):
            s = self.Positions[j, :]
            u = np.random.randn(len(s)) * sigma
            v = np.random.randn(len(s))
            step = u / abs(v) ** (1 / beta)
            stepsize = 0.01 * (step * (s - self.best_position))
            s = s + stepsize * np.random.randn(len(s))
            tempnest[j,:]=self.ensure_bounds(s)
            tempnest[j,:] = self.ensure_discrete(s)
         
        return tempnest
        
    def evolute(self, ngen, x0=None, verbose=False):
        """
        This function evolutes the CS algorithm for number of generations
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) initial position of the cuckoos (must be of same size as ``ncuckoos``)
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        self.history = {'local_fitness':[], 'global_fitness':[]}
        self.best_fitness=float("inf") 
        self.verbose=verbose
        self.Positions = np.zeros((self.ncuckoos, self.dim))
        if x0:
            assert len(x0) == self.ncuckoos, '--error: the length of x0 ({}) MUST equal the number of cuckoos in the group ({})'.format(len(x0), self.ncuckoos)
            for i in range(self.ncuckoos):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    self.Positions[i,:] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)
                else:
                    self.Positions[i,:] = x0[i]
        else:
            #self.Positions=self.init_sample(self.bounds)  #TODO, update later for mixed-integer optimisation
            # Initialize the positions of cuckoos
            for i in range(self.ncuckoos):
                self.Positions[i,:]=self.init_sample(self.bounds)

        fitness=self.eval_cuckoos() # evaluate the first cuckoos
        self.best_position, self.best_fitness = self.select(pos = self.Positions,fit = fitness) # find the initial best position and fitness
        for l in range(1, ngen+1):# Main loop
            
            self.a= 1 - (l-1) * ((1) / ngen)  #mir: a decreases linearly between 1 to 0, for discrete mutation
            #-----------------------------
            # Obtain new Cuckoo Positions by Lévy flights
            #-----------------------------
            newnest = self.UpdateCuckoos() # new cuckoos after Lévy Flights
            #----------------------
            #  Evaluate New Cuckoos
            #----------------------
            # Evaluating all new solutions
            tempnest = np.copy(self.Positions)
            fnew = self.eval_cuckoos(newnest)
            # and update current Cuckoos
            for j in range(0, self.ncuckoos):
                if fnew[j] <= fitness[j]:
                    fitness[j] = fnew[j]
                    tempnest[j, :] = newnest[j, :]
            self.Positions = tempnest.copy() # Take only the fittest individual
            #----------------------
            #  Discover a fraction ~pa of Cuckoos from the Cuckoos after Lévy flights
            #----------------------
            new_nest = np.zeros((self.ncuckoos, self.dim))
            K = np.random.uniform(0, 1, (self.ncuckoos, self.dim)) > self.pa
            stepsize = random.random() * (
                newnest[np.random.permutation(self.ncuckoos), :] - newnest[np.random.permutation(self.ncuckoos), :]
            )
            new_nest = newnest + stepsize * K # Update a fraction ~pa of Cuckoo after the Lévy flights
            
            #ensure bounds and discrete for the newnest
            for cuckoo in range(self.ncuckoos):
                new_nest[cuckoo,:]=self.ensure_bounds(new_nest[cuckoo,:])
                new_nest[cuckoo,:] = self.ensure_discrete(new_nest[cuckoo,:])
            #----------------------
            #  Re-evaluate the Cuckoos obtained and update to get the fittest individuals
            #----------------------
            tempnest = np.copy(self.Positions) # Will save the fittest individual
            fnew = self.eval_cuckoos(new_nest)
            for j in range(0, self.ncuckoos): # Compare Cuckoo fitness of newly generated Cuckoos (newnest) and current Cuckoos (self.positions)
                if fnew[j] <= fitness[j]:
                    fitness[j] = fnew[j]
                    tempnest[j, :] = new_nest[j, :]
            self.Positions = tempnest.copy() # Update the population
            
            #----------------------
            #  Logger related portion
            #----------------------
            for i, fits in enumerate(fitness):  
                #save the best of the best!!!
                if fits < self.best_fitness:
                    self.best_fitness=fits
                    self.best_position=self.Positions[i, :].copy()
            
            self.last_fit=np.array(fitness)
            self.last_pop=self.Positions.copy()
            
            #--mir
            if self.mode=='max':
                self.fitness_best_correct=-self.best_fitness
                self.local_fitness=-np.min(fitness)
            else:
                self.fitness_best_correct=self.best_fitness
                self.local_fitness=np.min(fitness)

            self.history['local_fitness'].append(self.local_fitness)
            self.history['global_fitness'].append(self.fitness_best_correct)
            # Print statistics
            if self.verbose and i % self.ncuckoos:
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('CS step {}/{}, ncuckoos={}, Ncores={}'.format((l)*self.ncuckoos, ngen*self.ncuckoos, self.ncuckoos, self.ncores))
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('Best Cuckoo/Nest Fitness:', np.round(self.fitness_best_correct,6))
                if self.grid_flag:
                    self.cuckoo_decoded = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
                    print('Best Cuckoo Position:', self.cuckoo_decoded)
                else:
                    print('Best Cuckoo Position:', self.best_position)
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')

        #mir-grid
        if self.grid_flag:
            self.cuckoo_correct = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
        else:
            self.cuckoo_correct = self.best_position   
        
        if self.mode=='max':
            self.last_fit=-self.last_fit
            
        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=False)
             
        if self.verbose:
            print('------------------------ CS Summary --------------------------')
            print('Best fitness (y) found:', self.fitness_best_correct)
            print('Best individual (x) found:', self.cuckoo_correct)
            print('--------------------------------------------------------------') 
    
        return self.cuckoo_correct, self.fitness_best_correct, self.history